from typing import Dict, Optional

import click

from globus_sdk import AuthClient

from gcs_cli.config import internal_auth_client, token_storage_adapter

GCS_SCOPE_FORMAT_STRING = scope = "urn:globus:auth:scope:{}:manage_collections"
BASE_SCOPES = (
    "openid profile email urn:globus:auth:scope:auth.globus.org:view_identity_set "
    "urn:globus:auth:scope:auth.globus.org:manage_projects "
)


def do_link_auth_flow(
    gcs_client_id: str,
    session_params: Optional[Dict] = None,
    force_new_client: bool = False,
) -> bool:
    """
    Prompts the user with a link to authenticate with globus auth
    and authorize the CLI to act on their behalf.
    """
    session_params = session_params or {}

    # get the ConfidentialApp client object
    auth_client = internal_auth_client(force_new_client=force_new_client)

    # start the Confidential App Grant flow
    auth_client.oauth2_start_flow(
        redirect_uri=auth_client.base_url + "v2/web/auth-code",
        refresh_tokens=True,
        requested_scopes=BASE_SCOPES + GCS_SCOPE_FORMAT_STRING.format(gcs_client_id),
    )

    # prompt
    additional_params = {"prompt": "login"}
    additional_params.update(session_params)
    linkprompt = "Please authenticate with Globus here"
    click.echo(
        "{0}:\n{1}\n{2}\n{1}\n".format(
            linkprompt,
            "-" * len(linkprompt),
            auth_client.oauth2_get_authorize_url(query_params=additional_params),
        )
    )

    # come back with auth code
    auth_code = click.prompt("Enter the resulting Authorization Code here").strip()

    # finish auth flow
    exchange_code_and_store_config(auth_client, auth_code, gcs_client_id)
    return True


def exchange_code_and_store_config(
    auth_client: AuthClient, auth_code: str, gcs_client_id: str
) -> None:
    """
    Finishes auth flow after code is gotten from command line or local server.
    Exchanges code for tokens and stores tokens using the configured storage
    adapter
    """
    adapter = token_storage_adapter()

    # do a token exchange with the given code
    tkn = auth_client.oauth2_exchange_code_for_tokens(auth_code)

    # revoke any existing tokens
    loaded = adapter.get_token_data(gcs_client_id) or {}
    if loaded:
        for token_key in ("access_token", "refresh_token"):
            if token_key in loaded:
                auth_client.oauth2_revoke_token(loaded[token_key])
        adapter.remove_tokens_for_resource_server(gcs_client_id)

    # store the new tokens in their place
    adapter.store(tkn)
