"""
A wrapper which nicely integrates SDK IdentityMaps into the gcs-cli

A "resolver" is defined with a field it uses in each item from the response data as the
full principal value. This is the "key".
Typical usage should be to define a resolver in each command which needs resolution, to
provide the key to use. Definition may have to happen outside of a command to
accommodate field lists which are defined as constants, and this is supported.

A resolver provides a field handler for principal URNs + a pagination callback to get
all principals from each page of results added to the ID map.
It can also be initialized to take identity IDs (not in Principal URN format).

This lets us get some of the benefit of the IdentityMap bulk calls without forcing the
whole paginated call to be walked at once.

This also lets us keep the resolution work only in the text-mode printed output (and not
applied on JSON output).
"""
from typing import Any, Mapping, Optional, Tuple

from globus_sdk import GlobusHTTPResponse, IdentityMap

import gcs_cli.services.auth as auth

IDENTITY_URN_PREFIX = "urn:globus:auth:identity:"


class PrincipalResolver:
    """
    Everything is done lazily via properties so that nothing happens during
    start up.

    Pass ``PrincipalResolver.field`` as a field key for output printing

    Pass ``PrincipalResolver.page_callback`` as a page callback for output printing

    Usage:

    >>> PrincipalResolver("urnfield")

    creates a resolver which pulls principal URNs from the field named "urnfield"

    >>> PrincipalResolver("idfield", use_urns=False)

    creates a resolver which pulls Identity IDs from the field named "idfield"
    """

    def __init__(self, key: Any, use_urns: bool = True):
        self.key = key
        self.use_urns = use_urns
        self._idmap: Optional[IdentityMap] = None

    @property
    def idmap(self) -> IdentityMap:
        if not self._idmap:
            self._idmap = IdentityMap(auth.get_client())
        return self._idmap

    def _raw_id_from_object(self, obj: Mapping) -> Tuple[bool, Any, Any]:
        """
        returns a 3-tuple, (resolved?, original, value)
        so that callers know if this helper thinks it got a good result or not
        """
        value = obj[self.key]
        # if not using URNs, the "raw ID" is just the value and it "always works"
        if not self.use_urns:
            return (True, value, value)

        # otherwise, check
        # if it doesn't have the URN prefix, it is not a valid URN, so this lookup
        # failed -- the result is (False, ...) to indicate failure
        if not value.startswith(IDENTITY_URN_PREFIX):
            return (False, value, value)
        # if it has the right prefix, left-strip the prefix as the new value, return the
        # original and the success indicator
        return (True, value, value[len(IDENTITY_URN_PREFIX) :])

    def field(self, data: Mapping) -> Any:
        resolved, original, value = self._raw_id_from_object(data)
        if not resolved:
            return value
        # try to do the lookup and get the "username" property
        # but default to the original value if this doesn't resolve
        return self.idmap.get(value, {}).get("username", original)

    def page_callback(self, data_page: GlobusHTTPResponse) -> None:
        for item in data_page.data["data"]:
            resolved, _original, value = self._raw_id_from_object(item)
            if resolved:
                self.idmap.add(value)


default_principal_resolver = PrincipalResolver("principal")
default_identity_id_resolver = PrincipalResolver("identity_id", use_urns=False)
