from typing import Optional, TypeVar, cast

import click

_T = TypeVar("_T")


class _State:
    def __init__(self) -> None:
        self.output_format: str = "text"
        self.explicit_host: Optional[str] = None
        self.show_stack_trace: bool = False


def format_option(f: _T) -> _T:
    def callback(ctx: click.Context, param: click.Parameter, value: str) -> None:
        if not value:
            return
        state = ctx.ensure_object(_State)
        state.output_format = value.lower()

    return cast(
        _T,
        click.option(
            "-F",
            "--format",
            type=click.Choice(["text", "json", "list"], case_sensitive=False),
            help="Output format for stdout. Defaults to text",
            expose_value=False,
            callback=callback,
        )(f),
    )


def explicit_host_option(f: _T) -> _T:
    def callback(ctx: click.Context, param: click.Parameter, value: str) -> None:
        if not value:
            return
        state = ctx.ensure_object(_State)
        state.explicit_host = value

    return cast(
        _T,
        click.option(
            "--use-explicit-host",
            envvar="GCS_CLI_EXPLICIT_HOST",
            expose_value=False,
            callback=callback,
            metavar="<ip_address>",
            help=("IP address of the GCS Node to use"),
        )(f),
    )


def show_stack_trace_option(f: _T) -> _T:
    def callback(ctx: click.Context, param: click.Parameter, value: str) -> None:
        if not value:
            return
        state = ctx.find_root().ensure_object(_State)
        state.show_stack_trace = value

    return cast(
        _T,
        click.option(
            "--show-stack-trace",
            hidden=True,
            envvar="GCS_CLI_SHOW_STACK_TRACE",
            is_flag=True,
            expose_value=False,
            callback=callback,
        )(f),
    )


def get_explicit_host() -> Optional[str]:
    ctx = click.get_current_context(silent=True)
    if ctx:
        state = cast(_State, ctx.ensure_object(_State))
        return state.explicit_host
    return None


def outformat_is_text() -> bool:
    ctx = click.get_current_context()
    state = cast(_State, ctx.ensure_object(_State))
    return state.output_format == "text"


def outformat_is_json() -> bool:
    ctx = click.get_current_context()
    state = cast(_State, ctx.ensure_object(_State))
    return state.output_format == "json"


def outformat_is_list() -> bool:
    ctx = click.get_current_context()
    state = cast(_State, ctx.ensure_object(_State))
    return state.output_format == "list"


def get_show_stack_trace() -> bool:
    ctx = click.get_current_context()
    state = cast(_State, ctx.ensure_object(_State))
    return state.show_stack_trace
