from flask import current_app, g, request
from functools import wraps

from globus.manager.api.errors import UnauthorizedError
from globus.manager.api.authentication.authstate import AuthenticationState
from globus.manager.api.authentication.utils import get_auth_token


def authenticated(fn):
    """Mark a route as requiring authentication."""

    @wraps(fn)
    def decorated_function(*args, **kwargs):
        if g.get("server_timing"):
            g.server_timing.start(("auth", "Authentication"))

        if not g.get("auth_state"):
            token = get_auth_token(request.headers)
            if not token:
                if "application/x-www-form-urlencoded" == request.mimetype:
                    token = request.form.get("access_token")
                    if not token:
                        raise UnauthorizedError()

                    if "Bearer " in token:
                        token = token[7:]
                else:
                    raise UnauthorizedError()

            g.auth_state = AuthenticationState(token)

            if not g.auth_state.is_authenticated:
                raise UnauthorizedError()

        if g.get("server_timing"):
            g.server_timing.finish(("auth", "Authentication"))

        return fn(*args, **kwargs)

    return decorated_function


def authenticated_with_secret(fn):
    """Mark a route as requiring basic authentication with a shared secret."""

    @wraps(fn)
    def decorated_function(*args, **kwargs):
        if g.get("server_timing"):
            g.server_timing.start(("auth", "Authentication"))

        if not request.authorization:
            raise UnauthorizedError()

        username = request.authorization.get("username")
        password = request.authorization.get("password")

        remote_addr = None
        remote_addr = request.environ.get("HTTP_X_FORWARDED_FOR")
        if remote_addr:
            remote_addr = remote_addr.split(",")[-1].strip()
        else:
            remote_addr = request.environ.get("REMOTE_ADDR")
        # REMOTE_ADDR is unset when contacted via unix domain socket.
        if remote_addr and remote_addr not in ["127.0.0.1", "::1"]:
            raise UnauthorizedError()

        if username != "gridftp":
            raise UnauthorizedError()
        if password not in current_app.gridftp_keys:
            raise UnauthorizedError()

        if g.get("server_timing"):
            g.server_timing.finish(("auth", "Authentication"))

        return fn(*args, **kwargs)

    return decorated_function
