"""
This module defines decorators which can be used to control access to various
GCS Manager API routes.
"""
from __future__ import annotations

from functools import partial, wraps
from typing import Callable, List, Optional, overload, TypeVar, Union
from typing_extensions import ParamSpec

from flask import current_app, g

from globus.manager.api.models import StorageGateway, UserCredential

from globus_sdk.scopes import GCSCollectionScopeBuilder, GCSEndpointScopeBuilder

import globus.manager.api.errors as errors
from .authentication_timeout_filter import AuthenticationTimeoutFilter
from .credential_owner import CredentialOwnerAuthorizer
from .identity_set import IdentitySet
from .public import PublicAuthorizer
from .role import (
    RoleAuthorizer,
    RoleSpec,
)
from .storage_gateway import StorageGatewayAuthorizer
from globus.manager.constants.roles import RoleType


T = TypeVar("T")
P = ParamSpec("P")


@overload
def requires_one_role(fn: Callable[P, T]) -> Callable[P, T]:
    ...


@overload
def requires_one_role(
    *,
    role_names: List[str] | None = None,
    require_session: Optional[bool] = None,
    authentication_timeout: Optional[Union[int, bool]] = False,
    require_mfa: Optional[bool] = None,
    storage_gateway_selector: Callable[[], Optional[StorageGateway]] | None = None,
    ignore_endpoint_admin_if_guest: bool = False,
    collection_selector: Callable | None = None,
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    ...


def requires_one_role(
    fn: Callable[P, T] | None = None,
    *,
    role_names: List[str] | None = None,
    require_session: Optional[bool] = None,
    authentication_timeout: Optional[Union[int, bool]] = False,
    require_mfa: Optional[bool] = None,
    ignore_endpoint_admin_if_guest: bool = False,
    storage_gateway_selector: Callable[[], Optional[StorageGateway]] | None = None,
    collection_selector: Callable | None = None,
):
    """
    Decorate a route to enforce a policy requiring the identity set of the
    caller have access to at least one of the listed roles.  If any of the
    roles is granted to the caller's identity set, decorator calls `fn`.

    :param callable fn:
        Function to call if the authorization requirements are met.
    :param List[str] role_names:
        A list of role string that can be processed with the
        `globus.manager.api.authorization.role.RoleSpec` constructor.
    :require_session:
        If True, roles are only considered if the caller's identity with
        that role has authenticated within the current request's Globus Auth
        session. By default, sessions are not required. If a value for this
        parameter is not provided, it will be derived from the policies of the
        storage gateway returned by calling the `storage_gateway_selector`
        or the policies of the storage gateway associated with the collection
        returned from calling the `collection_selector`.
    :authentication_timeout:
        If set, roles are only considered if the caller's identity with
        that role has authenticated within this timeout (in minutes). If
        `require_session` is True, then this filters based on session
        authentications, otherwise it filters based on last_authentications
        from any client application. If a value for this
        parameter is not provided, it will be derived from the policies of the
        storage gateway returned by calling the `storage_gateway_selector`
        or the policies of the storage gateway associated with the collection
        returned from calling the `collection_selector`.
    :require_mfa:
        If True, roles are only considered if the caller's identity with
        that role has authenticated within the current request's Globus Auth
        session using multi-factor authentication. By default, multi-factor
        authentication is not required. If a value for this
        parameter is not provided, it will be derived from the policies of the
        storage gateway returned by calling the `storage_gateway_selector`
        or the policies of the storage gateway associated with the collection
        returned from calling the `collection_selector`.
    :ignore_endpoint_admin_if_guest:
        If True, ignore endpoint:administrator roles for guest collections. This
        is used to prevent endpoint admins to gaining the ability to grant roles
        and acls on guest collections.
    :param callable storage_gateway_selector:
        A callable which will be called to resolve a storage_gateway associated
        with the current request.
    :param callable collection_selector:
        A callable which will be called to resolve a collection associated with
        the current request. This is required to evaluate collection
        roles but is not needed for endpoint roles.

    :raises MissingRequiredRoleError:
        The desired role is not available to the caller.
    """
    if fn is None:
        return partial(
            requires_one_role,
            role_names=role_names,
            require_session=require_session,
            authentication_timeout=authentication_timeout,
            require_mfa=require_mfa,
            storage_gateway_selector=storage_gateway_selector,
            ignore_endpoint_admin_if_guest=ignore_endpoint_admin_if_guest,
            collection_selector=collection_selector,
        )

    @wraps(fn)
    def decorated_function(*args: P.args, **kwargs: P.kwargs) -> T:
        collection = None
        role_specs = []
        if collection_selector:
            collection = collection_selector()
            if role_names:
                for role_name in role_names:
                    # Ignore collection-specific roles when no collection is
                    # selected
                    if role_name.endswith(":{collection_id}") and collection is None:
                        continue
                    # Ignore roles referring to the associated mapped collection
                    # if there is no collection or it is not a guest collection
                    if role_name.endswith(":{mapped_collection_id}") and (
                        collection is None or not collection.is_guest_collection
                    ):
                        continue
                    # Ignore collection roles when collection is deleted
                    if role_name.endswith(":{collection_id}") and collection.is_deleted:
                        continue
                    # Ignore roles referring to the associated mapped collection
                    # if the collection or the mapped collection is deleted.
                    if role_name.endswith(":{mapped_collection_id}") and (
                        collection.is_deleted or collection.mapped_collection.is_deleted
                    ):
                        continue
                    if (
                        ignore_endpoint_admin_if_guest
                        and collection
                        and collection.is_guest_collection
                        and role_name == RoleType.endpoint_administrator.value
                    ):
                        continue
                    role_specs.append(RoleSpec(role_name, collection))
        else:
            for role_name in role_names or []:
                role_specs.append(RoleSpec(role_name))

        role_authorizer = RoleAuthorizer(role_specs)
        g.authorizer = role_authorizer
        try:
            _ = g.auth_state
        except AttributeError:
            raise errors.UnauthorizedError()

        authorized_identities = role_authorizer.authorize_identities(g.auth_state)

        if authorized_identities:
            # Filter identities which have a required role based on the
            # required session and/or timeout information, if there is any.

            storage_gateway = None
            if storage_gateway_selector:
                # Override defaults if storage_gateway_selector is present
                storage_gateway = storage_gateway_selector()
                if not storage_gateway:
                    raise errors.NotFoundError()
            elif collection:
                # Override defaults if storage_gateway_selector is present
                storage_gateway = collection.storage_gateway

            if storage_gateway:
                current_app.logger.debug(
                    "authorizer using storage_gateway %s with require_session=%s, "
                    "authentication_timeout=%s and require_mfa=%s, passed in "
                    "require_session=%s, authentication_timeout=%s and require_mfa=%s",
                    storage_gateway,
                    storage_gateway.high_assurance,
                    storage_gateway.authentication_timeout_mins,
                    storage_gateway.require_mfa,
                    require_session,
                    authentication_timeout,
                    require_mfa,
                )

            if require_session is None:
                if storage_gateway:
                    authorizer_require_session: bool = storage_gateway.high_assurance  # type: ignore
                else:
                    authorizer_require_session = False
            else:
                authorizer_require_session = require_session

            if authentication_timeout is False:
                if storage_gateway:
                    authorizer_authentication_timeout = (
                        storage_gateway.authentication_timeout_mins
                    )
                else:
                    authorizer_authentication_timeout = None
            else:
                authorizer_authentication_timeout = authentication_timeout

            if require_mfa is None:
                if storage_gateway:
                    authorizer_require_mfa = storage_gateway.require_mfa
                else:
                    authorizer_require_mfa = False
            else:
                authorizer_require_mfa = require_mfa

            authentication_timeout_filter = AuthenticationTimeoutFilter(
                require_session=authorizer_require_session,
                authentication_timeout=authorizer_authentication_timeout,
                require_mfa=authorizer_require_mfa,
            )
            authorized_identities = authentication_timeout_filter(
                authorized_identities, g.auth_state
            )

            if authorized_identities:
                g.authorized_identities = authorized_identities
                assert fn is not None
                return fn(*args, **kwargs)
            else:
                # Caller has a role, but it's not in session
                g.authorized_identities = []

                if authorized_identities.error_detail:
                    raise errors.AuthenticationTimeoutError(
                        detail=authorized_identities.error_detail
                    )
                else:
                    raise errors.MissingRequiredRoleError(roles=role_specs)
        else:
            # No role is available
            g.authorized_identities = []
            raise errors.MissingRequiredRoleError(roles=role_specs)

    return decorated_function


@overload
def requires_credential_owner(fn: Callable[P, T]) -> Callable[P, T]:
    ...


@overload
def requires_credential_owner(
    *,
    credential_selector: Callable[[], Optional[UserCredential]] | None = None,
    allow_admin_if_configured: bool = False,
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    ...


def requires_credential_owner(
    fn: Callable[P, T] | None = None,
    *,
    credential_selector: Callable[[], Optional[UserCredential]] | None = None,
    allow_admin_if_configured: bool = False,
):
    """
    Decorate a route to enforce a policy requiring the identity set of the
    caller be the owner of a specific
    `globus.manager.api.models.UserCredential`.  If authorized,
    this credential, this decorator calls `fn`.

    :param callable fn:
        Function to call if the authorization requirements are met.
    :param callable credential_selector:
        A callable which will be called to resolve a credential associated with
        the current request. This is required to evaluate collection
        roles but is not needed for endpoint roles.
    :param bool allow_admin_if_configured:
        Flag indicating whether to fallback to role authorization if the
        admin_managed_credentials property is set on the owning storage
        gateway.

    :raises NotFoundError:
        The credential does not exist.
    :raises NotResourceOwnerError:
        The credential is not owned by the coller.
    :raises InternalServerError:
        The credential_selector is misconfigured.
    """
    if fn is None:
        return partial(
            requires_credential_owner,
            credential_selector=credential_selector,
            allow_admin_if_configured=allow_admin_if_configured,
        )

    @wraps(fn)
    def decorated_function(*args: P.args, **kwargs: P.kwargs) -> T:
        g.authorized_identities = IdentitySet()
        if not credential_selector:
            raise errors.InternalServerError()

        credential = credential_selector()

        if not credential:
            raise errors.NotFoundError()

        storage_gateway: StorageGateway = credential.storage_gateway

        credential_authorizer = CredentialOwnerAuthorizer(credential)
        authz_result = credential_authorizer.authorize_identities(g.auth_state)
        g.authorized_identities = authz_result

        if authz_result:
            g.authorizer = credential_authorizer
        elif allow_admin_if_configured and storage_gateway.admin_managed_credentials:
            try:
                requires_one_role(
                    role_names=[
                        RoleType.endpoint_administrator.value,
                        RoleType.endpoint_owner.value,
                    ],
                    storage_gateway_selector=lambda: storage_gateway,
                )(lambda: True)()
                assert g.authorized_identities
            except Exception as e:
                current_app.logger.error(
                    f"requires_credential_owner() role check yielded {e}"
                )
                # Restore original authorization result, so that the ownership
                # error with details is raised
                g.authorized_identities = authz_result

        if g.authorized_identities:
            assert fn is not None
            return fn(*args, **kwargs)
        else:
            raise errors.AuthorizationError(
                detail=getattr(g.authorized_identities, "error_detail", None)
            )

    return decorated_function


def requires_data_access_scope(fn: Callable[P, T]) -> Callable[P, T]:
    """
    Decorate a route to require the caller to have the `data_access` scope for
    the collection model stored in g.mapped_collection if it is not a high
    assurance collection. If the scope is not present and the collection is
    not high assurance, return a 401 error requiring new authorization with
    that scope as a dependent scope of the manage_collections scope.

    :param fn:
        The route handler to call after the scope check
    """

    @wraps(fn)
    def decorated_function(*args: P.args, **kwargs: P.kwargs) -> T:
        collection = g.mapped_collection

        if not collection:
            raise errors.BadRequestError("Bad or missing mapped_collection_id")

        if not collection.high_assurance:
            collection_scope_builder = GCSCollectionScopeBuilder(collection.id)
            endpoint_scope_builder = GCSEndpointScopeBuilder(
                current_app.config["CLIENT_ID"]
            )
            scope_builder = endpoint_scope_builder.make_mutable("manage_collections")
            scope_builder.add_dependency(
                collection_scope_builder.data_access, optional=True
            )
            scopes = [scope_builder]

            try:
                g.auth_state.dependent_token(
                    collection.id, required_scope=collection_scope_builder.data_access
                )
            except errors.MissingRequiredScopesError:
                raise errors.MissingRequiredScopesError(scopes=scopes)
        assert fn is not None
        return fn(*args, **kwargs)

    return decorated_function


@overload
def requires_storage_gateway_policy(fn: Callable[P, T]) -> Callable[P, T]:
    ...


@overload
def requires_storage_gateway_policy(
    *,
    storage_gateway_selector: Callable[[], Optional[StorageGateway]] | None = None,
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    ...


def requires_storage_gateway_policy(
    fn: Callable[P, T] | None = None,
    *,
    storage_gateway_selector: Callable[[], Optional[StorageGateway]] | None = None,
):
    """
    Decorate a route to enforce a policy requiring the identity set of the
    caller be allowed by the policies associated with a particular storage
    gateway. If an identity in the caller's identity set meets the
    requirements, this decorator calls `fn`.

    :param callable fn:
        Function to call if the authorization requirements are met.
    :param callable storage_gateway_selector:
        A callable which will be called to resolve a storage_gateway associated
        with the current request. This must return a
        `globus.manager.api.models.StorageGateway` if the request can be
        resolved to one, or None otherwise.

    :raises BadRequestError:
        The storage_gateway_selector did not return a StorageGateway object.
    :raises ForbiddenError:
        The caller is not authorized by the Storage Gateway policies.
    """
    if fn is None:
        return partial(
            requires_storage_gateway_policy,
            storage_gateway_selector=storage_gateway_selector,
        )

    @wraps(fn)
    def decorated_function(*args: P.args, **kwargs: P.kwargs) -> T:
        g.authorized_identities = []
        storage_gateway = None
        if storage_gateway_selector:
            storage_gateway = storage_gateway_selector()
        else:
            raise errors.InternalServerError()
        if storage_gateway:
            storage_gateway_authorizer = StorageGatewayAuthorizer(storage_gateway)
            g.authorizer = storage_gateway_authorizer
            authorized_identities = storage_gateway_authorizer.authorize_identities(
                g.auth_state
            )

            if authorized_identities:
                g.authorized_identities = authorized_identities
                assert fn is not None
                return fn(*args, **kwargs)
            else:
                raise errors.AuthorizationError(
                    detail=authorized_identities.error_detail
                )
        else:
            raise errors.BadRequestError()

    return decorated_function


def requires_endpoint_visibility(fn: Callable[P, T]) -> Callable[P, T]:
    """
    Decorate a route to enforce a policy requiring the endpoint to be visible
    to an identity in the caller's identity_set.  An endpoint
    is visible to the caller if any of the following are true:

    -   The endpoint document's `public` property is True
    -   The caller has any role assignment for the endpoint or any
        collection on the endpoint
    -   The caller has an ACL granting access to any guest collection on
        this endpoint.
    -   The caller has an identity allowed by any storage gateway policy
        on this endpoint.

    :param callable fn:
        Function to call if the authorization requirements are met.

    :raises ForbiddenError:
        The caller is not authorized to view the endpoint.
    """

    @wraps(fn)
    def decorated_function(*args: P.args, **kwargs: P.kwargs) -> T:
        public_authorizer = PublicAuthorizer()
        authorized = public_authorizer.authorize_identities(g.auth_state)

        if authorized:
            g.authorizer = public_authorizer
        else:
            role_authorizer = RoleAuthorizer(
                [
                    "endpoint:owner",
                    "endpoint:administrator",
                    "endpoint:activity_manager",
                    "endpoint:activity_monitor",
                    "collection:administrator:*",
                    "collection:activity_manager:*",
                    "collection:activity_monitor:*",
                    "collection:access_manager:*",
                ]
            )

            authorized = role_authorizer.authorize_identities(g.auth_state)
            if authorized:
                g.authorizer = role_authorizer

        # TODO: ACL authorizer check
        if not authorized:
            for sg in StorageGateway.query.all():
                sg_authorizer = StorageGatewayAuthorizer(sg)
                authorized = sg_authorizer.authorize_identities(g.auth_state)
                if authorized:
                    g.authorizer = sg_authorizer
                    break

        if authorized:
            g.authorized_identities = authorized
            assert fn is not None
            return fn(*args, **kwargs)
        else:
            g.authorized_identities = []
            if authorized.error_detail:
                raise errors.AuthorizationError(
                    error_description="This endpoint is not "
                    "visible to any of your identities",
                    detail=authorized.error_detail,
                )
            else:
                raise errors.AuthorizationError(
                    error_description="This endpoint is not "
                    "visible to any of your identities",
                )

    return decorated_function


@overload
def requires_subscription(fn: Callable[P, T]) -> Callable[P, T]:
    ...


@overload
def requires_subscription(
    *, level: Optional[str] = None
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    ...


def requires_subscription(
    fn: Callable[P, T] | None = None, *, level: Optional[str] = None
):
    """
    Decorate a route to require a subscription for the route to be
    considered accessible. If no subscription is present on the endpoint,
    raise the subscription error. This is intended to be used before
    role based authorization checks in order to make the errors more
    understandable.

    :param fn:
        The route handler to call after the subscription check.
    :param level:
        The minimum subscription level to use. If this is not supplied by the
        caller, `starter` is used.
    """
    if fn is None:
        return partial(requires_subscription, level=level)

    if level is None:
        level = "starter"

    @wraps(fn)
    def decorated_function(*args: P.args, **kwargs: P.kwargs) -> T:
        if not current_app.subscription and level != "basic":
            raise errors.SubscriptionRequiredError(subscription_level=level)
        assert fn is not None
        return fn(*args, **kwargs)

    return decorated_function
