from typing import List, Mapping

from flask import current_app

import globus.manager.api.errors as errors
from .authentication_timeout_filter import AuthenticationTimeoutFilter
from .identity_set import IdentitySet
from .identity_in_identity_set_filter import IdentityInIdentitySetFilter


class RequiredDomainFilter:
    def __init__(self, required_domains):
        self.required_domains = [d.lower() for d in required_domains]

    def __call__(self, identity_set: IdentitySet, auth_state):
        """
        Filter the identity_set, removing all items that do not have
        a username in any of the required_domains of this filter.

        :param identity_set:
            Information about Auth identities.
        :param auth_state:
            Authentication state for this request

        :returns: Returns the subset of identity_set_data which match
            any of the required_domains allowed by this filter.
        """
        invalid_identities: List[Mapping] = []

        if len(self.required_domains) > 0:
            invalid_identities = [
                i
                for i in identity_set
                if i.get("username", "@").rsplit("@", 1)[1].lower()
                not in self.required_domains
            ]
            for i in invalid_identities:
                identity_set.remove(i)

        if invalid_identities:
            detail = errors.NotFromAllowedDomain(self.required_domains)
            identity_set.error_details.insert(0, detail)

        current_app.logger.debug("%s result=%s", self, identity_set)
        return identity_set

    def __repr__(self):
        return "<RequiredDomainFilter {}>".format(self.required_domains)


class StorageGatewayAuthorizer:
    """
    This class provides methods to process the storage gateway authorization
    policies
    """

    def __init__(self, storage_gateway, enforce_authentication_timeout=True):
        """
        Initialize an authorizer for a Storage Gateway. This filters
        on allowed domains and authentication time.

        :param storage_gateway:
            Storage Gateway object containing policies to enforce.
        :param enforce_authentication_timeout:
            Boolean flag whether to enforce authentication timeout or not.
        """
        self.filters = []
        self.required_domains = storage_gateway.allowed_domains

        # Filter on domains first so the more obvious error will occur
        if len(self.required_domains) > 0:
            self.filters.append(RequiredDomainFilter(self.required_domains))

        if enforce_authentication_timeout:
            self.filters.append(
                AuthenticationTimeoutFilter(
                    require_session=storage_gateway.high_assurance,
                    authentication_timeout=(
                        storage_gateway.authentication_timeout_mins
                    ),
                    require_mfa=getattr(storage_gateway, "require_mfa", False),
                )
            )

    def authorize_identity(self, identity_id: str, auth_state):
        """
        Return an entry from identity_set_data which matches the policies
        for this storage gateway.  Error string should include
        'current session' to trigger a login in the collections app.
        """
        current_app.logger.debug(
            "Attempting to authorize %s as %s",
            auth_state.identity_set_data,
            identity_id,
        )

        idset = IdentitySet(identity_set=auth_state.identity_set_data)

        # Filter out identities which don't match the requested id first
        f = IdentityInIdentitySetFilter(identity_id)
        f(idset, auth_state)

        for f in self.filters:
            f(idset, auth_state)

        if idset:
            assert len(idset) == 1
            return idset[0]

        detail = None
        if idset.error_details:
            detail = idset.error_details[0]
        raise errors.AuthorizationError(detail=detail)

    def authorize_identities(self, auth_state):
        """
        Return a list of entries from identity_set_data which match the
        storage gateway policies.
        """
        details = []
        idset = IdentitySet()
        for i in auth_state.identity_set_data:
            try:
                self.authorize_identity(i["id"], auth_state)
                idset.append(i)
            except errors.APIBaseError as e:
                if e.detail:
                    details.append(e.detail)
                current_app.logger.debug(e)

        if not idset:
            if not details:
                details = [errors.NotFromAllowedDomain(self.required_domains)]
            idset.error_details = details

        return idset
