from __future__ import annotations

import enum
from typing import List
import datetime
import uuid

from flask import current_app, g, jsonify

import globus.manager.constants as c
from globus.manager.api.schemas import data_type_schema as data_types
from globus.manager.api.authentication.authstate import AuthenticationState
from globus.manager.api.authentication.decorators import authenticated_with_secret
from globus.manager.api.authorization import StorageGatewayAuthorizer
from globus.manager.api.blueprints.decorators import (
    enum_list_converter,
    expects_json,
    query_parameter,
    result_data,
)
from globus.manager.api.blueprints.utils import (
    check_subscription,
    create_object,
    put_object,
    update_object,
)
import globus.manager.api.models as m
import globus.manager.api.errors as errors
from globus.manager.api.models.result import Result
from .openapi import OpenAPIBlueprint


blueprint = OpenAPIBlueprint(
    "private_gridftp_blueprint",
    __name__,
    tag="Private GridFTP API",
    schema_name="private-gridftp.yaml",
    description="Private GridFTP API",
    url_prefix="/api",
)


class LoginInclude(enum.Enum):
    collection = "collection"
    storage_gateway = "storage_gateway"
    user_credential = "user_credential"


@blueprint.route(
    "/private/gridftp_login/<uuid:collection_id>",
    methods=["POST"],
    summary="Process login information",
    private=True,
)
@result_data(
    description="Login response",
    schemas=[
        data_types.Collection,
        data_types.StorageGateway,
        data_types.UserCredential,
    ],
)
@query_parameter(
    name_spec="include",
    description="""
    List of documents to return in the response. This may consist of any of the
    following values: **collection**, **storage_gateway**, and **user_credential**.
    """,
    converter=enum_list_converter(LoginInclude),
    default=[],
)
@authenticated_with_secret
@expects_json(
    schema=data_types.GridFTPLogin, description="Login information", required=True
)
def private_gridftp_login_post(
    collection_id: uuid.UUID,
    include: List[LoginInclude],
) -> None:
    """
    Process the passed in identity_set to determine the login information for
    the specified collection. The "include" parameter determines which
    documents are included in the response.

    """
    username: str | None = g.validated_data.get("username")
    collection = m.Collection.query.filter(
        m.Collection.id == str(collection_id)
    ).one_or_none()

    if not collection:
        raise errors.NotFoundError(
            "The collection {} does not exist.".format(str(collection_id))
        )
    elif collection.is_deleted:
        raise errors.NotFoundError(
            "The collection {} has been deleted.".format(str(collection_id))
        )

    if collection:
        check_subscription(
            collection.storage_gateway.connector,
            high_assurance=collection.storage_gateway.high_assurance,
            collection_type=c.CollectionType(collection.collection_type),
        )

        if collection.is_guest_collection:
            if not collection.mapped_collection.allow_guest_collections:
                raise errors.NotFoundError("No such collection")

    storage_gateway = collection.storage_gateway
    user_credential = None

    if storage_gateway.deleted:
        raise errors.NotFoundError()

    current_app.logger.info(
        "login: %s identities: %s", collection, g.validated_data["identity_ids"]
    )

    identity_set: list[str] = []

    if c.CollectionType(collection.collection_type) == c.CollectionType.guest:
        # For a guest collection, we will process the API based on the
        # identity_id information in the collection's user_credential if
        # present.
        if collection.user_credential:
            identity_set = [collection.user_credential.identity_id]
        else:
            identity_set = [collection.identity_id]
    elif c.CollectionType(collection.collection_type) == c.CollectionType.mapped:
        # For a mapped collection, we process the API based on the
        # identity_id information passed to this API.
        if g.validated_data["identity_ids"]:
            identity_set = [str(iid) for iid in g.validated_data["identity_ids"]]

    # Synthesize some authentication state based on the identities
    # we've received from the GridFTP server. Those were verified
    # by Transfer at the time of the job submission, or by HTTPS
    # by introspecting the auth tokens.
    auth_state = AuthenticationState(identity_set=identity_set)

    # Because the authentication timeouts were validated at the
    # time the job was submitted from transfer, we don't verify
    # them now.
    authorizer = StorageGatewayAuthorizer(
        storage_gateway, enforce_authentication_timeout=False
    )
    authorized_identities = authorizer.authorize_identities(auth_state)

    if not authorized_identities:
        detail = None
        if authorized_identities.error_details:
            detail = authorized_identities.error_details[0]
        raise errors.AuthorizationError(detail=detail)

    # Map the authorized identity set to allowed local usernames if any
    mappings = collection.storage_gateway.map_identities(authorized_identities)

    if mappings and username:
        # If gridftp requested a specific local user, filter accounts to be
        # those that match that username
        mappings = list(filter(lambda account: account.username == username, mappings))

    if not mappings:
        if username:
            raise errors.InvalidUserError(
                error_description="No mapping exists from your identity "
                f'set to the requested account "{username}" on collection '
                f"{collection.id}",
                usernames=[username],
            )
        else:
            raise errors.InvalidUserError()

    for map_result in mappings:
        if collection.user_credential:
            # For a guest collection, the collection will have a
            # specific user_credential associated with it. If it matches
            # our identity we can use it
            if collection.user_credential.identity_id == map_result.identity_id:
                if collection.user_credential.username == map_result.username:
                    user_credential = collection.user_credential
                    break
        else:
            # For a mapped collection, we will see if the user has
            # registered a user_credential on the storage_gateway. If the
            # Storage Gateway maps identities, then make sure the
            # credential maps to that identity, otherwise, choose the
            # credential that is registered
            user_credential = m.UserCredential.query.filter(
                ~m.UserCredential.deleted,  # type: ignore
                m.UserCredential.storage_gateway_id == storage_gateway.id,
                m.UserCredential.identity_id == map_result.identity_id,
                m.UserCredential.username == map_result.username,
            ).one_or_none()
        if user_credential:
            break

    if collection.user_credential and not user_credential:
        # This collection has a user credential, but none of the mappings are
        # for that credential. This means that an identity has changed since
        # the credential was created and it is no longer usable
        raise errors.InvalidCredentialError(
            "This collection was created with a credential which is "
            "no longer accessible by the account which created it",
            user_credential_id=collection.user_credential.id,
        )

    if collection.user_credential and collection.is_guest_collection:
        if not collection.mapped_collection.check_sharing_user_allowed(
            collection.user_credential.username
        ):
            raise errors.InvalidUserError(
                "This guest collection was created with an account "
                "which is no longer permitted to have guest collections.",
                usernames=[collection.user_credential.username],
            )

    if user_credential and not user_credential.invalid:
        if user_credential.needs_refresh():
            refresh_successful = False
            try:
                user_credential.refresh()
                refresh_successful = True
            except Exception as e:
                # If refresh fails, credential may have been marked invalid
                current_app.logger.debug("Credential refresh failed: %s", e)
            try:
                update_object(user_credential, lambda o: o)
            except Exception as e:
                current_app.logger.warning(
                    "Error refreshing %s: %s", user_credential, str(e)
                )
                m.db.session.rollback()
                refresh_successful = False

            if not refresh_successful:
                raise errors.InvalidCredentialError(
                    "Unable to refresh user credential",
                    user_credential_id=user_credential.id,
                )

        # Raises an error if the credential can not be used for this collection
        try:
            collection.check_user_credential(user_credential)
        except Exception as e:
            if user_credential.invalid:
                update_object(user_credential, lambda o: o)
            raise e

    data = []

    if not user_credential or user_credential.invalid:
        if user_credential:
            deleted_user_credential = user_credential
            authorized_identity = next(
                ai
                for ai in authorized_identities
                if ai["id"] == user_credential.identity_id
            )
            username = user_credential.username
        elif len(mappings) == 1 or (
            len(mappings) > 1 and not collection.storage_gateway.require_credential
        ):
            # Webapp doesn't support managing anon s3 creds, so if there are
            # multiple mappings and its an anon s3 endpoint, just use the first
            # which meets domain requirements.
            authorized_identity = next(
                ai
                for ai in authorized_identities
                if ai["id"] == mappings[0].identity_id
            )
            # If there's an existing credential, we'll need to update
            # instead of create in appsync
            deleted_user_credential = m.UserCredential.query.filter(
                m.UserCredential.deleted,
                m.UserCredential.storage_gateway_id == storage_gateway.id,
                m.UserCredential.identity_id == authorized_identity["id"],
                m.UserCredential.username == mappings[0].username,
            ).one_or_none()
        elif collection.storage_gateway.require_credential:
            raise errors.CredentialNotFoundError(
                accounts=[mapping for mapping in mappings]
            )

        # If there is no user credential in the db, and there is only
        # one mapping for the user, create one on the fly. This may be
        # marked as an invalid credential for connectors that require
        # some intervention to populate the credential
        uc_doc = {
            "DATA_TYPE": "user_credential#1.0.0",
            "identity_id": authorized_identity["id"],
            "username": mappings[0].username,
            "storage_gateway_id": storage_gateway.id,
        }
        if deleted_user_credential and collection.is_guest_collection:
            # Don't revive deleted credentials on guest access
            user_credential = deleted_user_credential
        elif deleted_user_credential:
            # For some connectors with service creds, a refresh could fail because of
            # an issue with the service cred. In that case it's set as invalid
            # but not deleted. This is a funny state. Clear that if the user tries
            # the cred again, and hopefully we can reuse it
            if deleted_user_credential.invalid and not getattr(
                collection.storage_gateway, "user_credential_required", True
            ):
                deleted_user_credential.invalid = False
            deleted_user_credential.deleted = False
            put_object(deleted_user_credential, uc_doc)
            user_credential = deleted_user_credential
        else:
            cls = m.UserCredential.polymorphic_class(storage_gateway.connector_id)
            current_app.logger.error("Creating new credential object from %s", uc_doc)
            user_credential = cls(
                doc=uc_doc,
                storage_gateway=storage_gateway,
            )
            current_app.logger.error(
                "Creating new credential object %s", user_credential
            )
            create_object(user_credential)

    if not user_credential:
        raise errors.CredentialNotFoundError(accounts=[mapping for mapping in mappings])
    elif user_credential.invalid:
        if c.CollectionType(collection.collection_type) == c.CollectionType.guest:
            raise errors.InvalidCredentialError(
                "This collection is unusable, as it is configured with an invalid "
                "credential. Please contact the collection owner to update "
                "its configuration."
            )
        elif user_credential.provisioned:
            raise errors.InvalidCredentialError(user_credential_id=user_credential.id)
        else:
            raise errors.InvalidCredentialError(
                "Your credential requires some initial setup.",
                user_credential_id=user_credential.id,
            )

    now = datetime.datetime.now(datetime.timezone.utc)
    last_access = now.date()

    if collection.last_access is None or last_access > collection.last_access:
        update_last_access = {"last_access": str(last_access)}
        # Check if we're planning on doing this already
        pending_update = m.PendingUpdate.query.filter(
            m.PendingUpdate.object_id == collection.id,
            m.PendingUpdate.update_type == m.PendingUpdate.UPDATE_LAST_ACCESS,
        ).one_or_none()

        if pending_update:
            try:
                pending_last_access = datetime.datetime.strptime(
                    pending_update.data["last_access"], "%Y-%m-%d"
                ).date()
            except Exception as e:
                current_app.logger.error(
                    "Internal inconsistency: unable to parse pending "
                    f"last_access: {e}"
                )
            else:
                if pending_last_access < last_access:
                    pending_update.data = update_last_access
                    try:
                        m.db.session.add(pending_update)
                        m.db.session.commit()
                    except Exception as e:
                        current_app.logger.warning(
                            "Internal inconsistency: unable to commit "
                            f"update to last_access: {e}"
                        )
        else:
            pending_update = m.PendingUpdate(
                object_id=collection.id,
                change_time=now,
                endpoint_serial=collection.endpoint.serial,
                collection_serial=collection.serial,
                update_type=m.PendingUpdate.UPDATE_LAST_ACCESS,
                data=update_last_access,
            )
            m.db.session.add(pending_update)
            try:
                m.db.session.commit()
            except Exception as e:
                current_app.logger.warning(
                    "Unable to commit pending_update for last_access change "
                    f"for {collection.id}: {e}"
                )

    username = user_credential.username

    if LoginInclude.user_credential in include:
        data.append(data_types.UserCredential.private_schema.dump(user_credential))

    if LoginInclude.collection in include:
        collection_data = data_types.Collection.private_schema.dump(collection)
        if "sharing_restrict_paths" in collection_data:
            collection_data[
                "sharing_restrict_paths"
            ] = data_types.PathRestrictions.dump(
                collection.absolute_sharing_restrict_paths(username)
            )
        collection_data["root_path"] = collection.absolute_root_path(username)
        data.append(collection_data)

    if LoginInclude.storage_gateway in include:
        storage_gateway_data = data_types.StorageGateway.private_schema.dump(
            storage_gateway
        )
        if "restrict_paths" in storage_gateway_data:
            storage_gateway_data["restrict_paths"] = data_types.PathRestrictions.dump(
                storage_gateway.absolute_restrict_paths(username=username)
            )

        data.append(storage_gateway_data)

    res = data_types.Result.dump(Result(message="Login ok", data=data))

    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response
