from flask import current_app
import re
from typing import Any, Callable, Dict, List, Optional

from sqlalchemy.exc import IntegrityError

import globus_sdk

from globus.manager.api.models import (
    Collection,
    Connector,
    ConfigurationObject,
    Endpoint,
    PosixStorageGateway,
    db,
)
from globus.manager.appsync.query import GraphQLError
import globus.manager.api.errors as errors
from globus.manager.constants import CollectionType


def check_subscription(
    connector: Connector,
    high_assurance: bool = False,
    collection_type: Optional[CollectionType] = None,
):
    """
    Perform a subscription check to create a resource using the given
    storage gateway and/or collection type. If either are missing, they
    do not contribute to the required subscription level and addons.

    By default, endpoint operations requires only the "basic" level of
    subscription, but more is required for the following cases:

    -   Guest collections require at least a starter level subscription
    -   Premium connectors require at least a standard level subscription
    -   High assurance storage gateways and collections require at least a
        high_assurance level subscription

    Additionally, when a user attempts to create a resource using a premium
    connector, they will need an add-on for the connector in addition to the
    standard subscription.

    :param connector:
        The connector being checked against the endpoint's subscription.
    :param high_assurance:
        An optional flag indicating whether this check is being done to
        create a high assurance storage gateway. This value is only checked
        if the `storage_gateway` parameter is a class type.
    :param collection_type:
        An optional collection type enum value indicating whether this is
        evaluating the ability to access or create a mapped or a guest
        collection.

    :raises errors.SubscriptionRequiredError:
        If the storage_gateway or collection type indicate a non-basic
        level of subscription is needed, this function raises the
        SubscriptionRequiredError which will indicate what features are
        required in the subscription.
    """
    subscription_level = "basic"
    addons = []

    if collection_type == CollectionType.guest:
        subscription_level = "starter"

    if connector.id != PosixStorageGateway.CONNECTOR_ID:
        addons.append(connector.display_name)
        subscription_level = "standard"

    if high_assurance:
        subscription_level = "standard"
        addons.append("high_assurance")

    # XXX: Add support for subscription service to check properties of the
    # subscription. Right now, assume a subscription allows everything
    if not current_app.subscription and subscription_level != "basic":
        raise errors.SubscriptionRequiredError(
            subscription_level=subscription_level, add_ons=addons
        )


def update_object(
    obj: ConfigurationObject,
    update_callable: Callable,
    *args: Any,
    dependent_objs: Optional[List[ConfigurationObject]] = None,
    **kwargs: Any,
) -> None:
    """
    Update the given object using update_callable (called with `obj`, `*args`,
    `**kwargs`), then update it in the database and appsync.
    """
    committed_to_appsync = False
    retries_remaining = 5

    with db.session.no_autoflush:
        while not committed_to_appsync and retries_remaining > 0:
            try:
                update_callable(obj, *args, **kwargs)
                db.session.add(obj)
            except (ValueError, KeyError, AttributeError) as e:
                current_app.logger.error("UPDATE1: %s %s %s", str(obj), type(e), str(e))
                obj.rollback()
                db.session.rollback()
                raise errors.ValidationError(str(e))
            except errors.APIBaseError as e:
                current_app.logger.error("UPDATE2: %s %s %s", str(obj), type(e), str(e))
                obj.rollback()
                db.session.rollback()
                raise
            except Exception as e:
                current_app.logger.error("UPDATE3: %s %s %s", str(obj), type(e), str(e))
                db.session.rollback()
                obj.rollback()
                raise errors.InternalServerError(str(e))

            try:
                current_app.appsync_client.update_object(
                    obj, dependent_objs=dependent_objs
                )
                committed_to_appsync = True
            except GraphQLError as e:
                current_app.logger.warning(
                    "UPDATE4: %s: %s", str(obj), e.message
                )  # noqa

                error_types = list(map(lambda e: e.get("errorType"), e.errors))
                if "NOT_FOUND" in error_types:
                    # One or more items do not exist. In the case of
                    # dependent_objs, we won't even be able to tell which one
                    # is missing. Just rollback and error.
                    db.session.rollback()
                    obj.rollback()
                    raise errors.NotFoundError()

                # Bad serial(s) or the object has been deleted.
                if "CONFLICT" in error_types:

                    def _find_item(items, obj_id):
                        for item in items:
                            if item["id"] == obj_id:
                                return item
                        return {}

                    db.session.rollback()
                    obj.rollback()

                    dep_error_info = None
                    obj_error_info = _find_item(e.data["items"], obj.id)
                    if dependent_objs:
                        for do in dependent_objs:
                            dep_error_info = _find_item(e.data["items"], do.id)

                            if (
                                dep_error_info
                                and dep_error_info.get("deleted")
                                or obj_error_info.get("deleted")
                            ):
                                raise errors.NotFoundError()

                            if dep_error_info:
                                do.update(dep_error_info)
                                # Commit here because the pending change to obj
                                # may conflict with this update and we want to
                                # save this update.
                                try:
                                    db.session.commit()
                                except IntegrityError:
                                    db.session.rollback()
                                    raise errors.ConflictError(
                                        "Unable to synchronize object {obj} locally"
                                    )

                    if obj_error_info:
                        obj.rollback({"items": [obj_error_info]})

                # Treat other errors types as other GraphQL or transmission errors
                retries_remaining -= 1

            except Exception as e:
                # Unexpected error
                db.session.rollback()
                obj.rollback()
                current_app.logger.error(
                    "UPDATE5 %s: %s %s:", str(obj), str(type(e)), str(e)
                )
                raise errors.BadGatewayError()

        if not committed_to_appsync:
            db.session.rollback()
            obj.rollback()
            raise errors.ConflictError(
                "Unable to synchronize object {} with other nodes".format(obj)
            )

        try:
            db.session.commit()
        except IntegrityError:
            db.session.rollback()
            raise errors.ConflictError(
                "Unable to synchronize object {} locally".format(obj)
            )


def patch_object(
    obj: ConfigurationObject,
    patch_document: Dict[str, Any],
    dependent_objs: Optional[List[ConfigurationObject]] = None,
) -> None:
    """
    Patch the given object, update it in the database and appsync.
    """
    return update_object(
        obj, lambda o: o.patch(**patch_document), dependent_objs=dependent_objs
    )


def put_object(
    obj: ConfigurationObject,
    put_document: Dict[str, Any],
    dependent_objs: Optional[List[ConfigurationObject]] = None,
) -> None:
    """
    Put the given object, update it in the database and appsync.
    """
    return update_object(
        obj, lambda o: o.put(**put_document), dependent_objs=dependent_objs
    )


def create_object(obj: ConfigurationObject) -> None:
    """
    Publish the given object to appsync and commit it to the database
    """
    committed_to_appsync = False
    retries_remaining = 5

    while not committed_to_appsync and retries_remaining > 0:
        try:
            current_app.appsync_client.create_object(obj)
            committed_to_appsync = True
        except GraphQLError as e:
            current_app.logger.warning("POST: %s: %s", obj.id, e.message)  # noqa
            if e.errors[0].get("errorType") == "CONFLICT":
                obj.rollback()
                db.session.rollback()
                raise errors.ResourceExistsError()
            else:
                retries_remaining -= 1
        except globus_sdk.GlobusAPIError as e:
            current_app.logger.warning("POST: %s: %s", obj.id, e.message)  # noqa
            retries_remaining -= 1
        except Exception as e:
            current_app.logger.error("POST: %s: %s", str(obj), str(e))
            break

    if not committed_to_appsync:
        current_app.logger.error(
            "POST: %s: failed to create object in appsync", str(obj)
        )
        obj.delete()
        db.session.commit()
        raise errors.BadGatewayError()

    db.session.add(obj)
    db.session.commit()


def delete_object(
    obj: ConfigurationObject,
    *args: Any,
    update: bool = False,
    **kwargs: Any,
) -> None:
    """
    Delete the given object and remove it from the database and appsync.
    There are two modes for this function: a soft delete (which marks the
    object as deleted but does not remove it from appsync) and a hard delete
    (which completely removes the record from appsync).

    The soft delete is needed in any case where the object is referred to by
    a foreign key in another object table. For example, collections refer
    to the storage gateway they are created from.

    A hard delete is for leaf objects or those where it is known that all
    objects which depend on them have been already deleted. This code assumes
    the caller chooses the correct one by setting the update flag to True for
    a soft delete.

    Deletion is performed in two steps: first this updates the appsync record
    to mark it as deleted, and second delete the resources used by the object.
    This order is chosen because the resource deletion in transfer/auth/dns can
    not be rolled back, but the appsync object can be recreated if there was a
    problem deleting the resources.

    There is a possibility that an object will be marked as deleted
    in appsync, but because of an error or crash, the underlying resources in
    transfer/auth/dns are not deleted. The object will be considered deleted
    from the GCS API perspective, so operations trying to use it will fail and
    it will not be returned from the API. Depending on the type of object,
    we might be able to clean it up out of band (manually), but there is
    currently not automated garbage collection.

    The parameters *args and **kwargs are additional arguments and named
    parameters which are passed to the object's delete method when it is called.
    """
    tag = obj.tag
    obj_id = obj.id

    # Save the state of the object in case the appsync delete succeeds but the
    # actual resource deletion fails
    rollback_data = obj.raw_configuration()
    obj_deleted = obj.deleted
    object_exists = True

    def _rollback(o):
        o.deleted = obj_deleted
        o.configuration = rollback_data

    committed_to_appsync = False

    # Try to delete (or soft-delete) the object in appsync
    retries_remaining = 5

    while not committed_to_appsync and retries_remaining > 0:
        if update:
            try:
                current_app.logger.debug(f"DELETING (soft) {obj.id}")
                obj.deleted = True
                current_app.appsync_client.update_object(obj)
                committed_to_appsync = True
            except GraphQLError as e:
                current_app.logger.error(
                    "DELETE (soft) %s: %s", obj.id, e.message
                )  # noqa
                if e.error_type == "NOT_FOUND":
                    # The object does not exist, either because it was never sent
                    # to appsync or was previously deleted and the TTL expired
                    #
                    # We should consider deletion successful, if we need
                    # to rollback, we will have to create the object, as updates
                    # will fail.
                    current_app.logger.error(
                        "DELETE (soft) NOT FOUND: %s", obj.id
                    )  # noqa
                    committed_to_appsync = True
                    object_exists = False
                elif e.error_type == "CONFLICT":
                    # Conflict can be either
                    # * Object serial number changed
                    # * Object has deleted=True set in the appsync record. This
                    #   only happens when the ObjectDelete mutation has been completed
                    #   for this object.
                    #
                    # If we receive data in the response, then we can determine
                    # if the object was deleted in appsync or has a serial conflict.
                    # If we do not, we assume the object is deleted.
                    current_app.logger.error(
                        "DELETE (soft) CONFLICT: %s", obj.id
                    )  # noqa
                    conflict_result: Optional[Dict[str, Any]] = None
                    if e.data and e.data["items"]:
                        match = [i for i in e.data["items"] if i["id"] == obj_id]
                        if match:
                            conflict_result = match[0]

                    if conflict_result and conflict_result.get("deleted", False):
                        # Object is marked as deleted in appsync but the TTL
                        # has not yet expired, treat it as if the object is deleted
                        committed_to_appsync = True
                        object_exists = False
                    elif conflict_result:
                        # Object exists, but serial number conflict prevents the
                        # current update. Update the object to match the response
                        # and retry
                        obj.rollback(e.data)
                        obj.deleted = obj_deleted
                        retries_remaining -= 1
                    else:
                        # No object means it was deleted?
                        committed_to_appsync = True
                        object_exists = False
                else:
                    # Some other transient error
                    current_app.logger.error(
                        "DELETE (soft) TRANSIENT: %s %s", obj.id, str(e)
                    )  # noqa
                    _rollback(obj)
                    db.session.rollback()
                    retries_remaining -= 1
            except Exception as e:
                current_app.logger.error(
                    "DELETE (soft) non-GraphQL exception {} {}: {}".format(
                        tag, obj.id, str(e)
                    )
                )
                _rollback(obj)
                db.session.rollback()
                raise errors.BadGatewayError()
        else:
            try:
                current_app.appsync_client.delete_object(obj)
                object_exists = False
                committed_to_appsync = True
            except GraphQLError as e:
                current_app.logger.warning(
                    "DELETE (hard) %s: %s", obj.id, e.message
                )  # noqa
                if e.error_type == "CONFLICT":
                    # ObjectDelete does not check serial number, but currently
                    # returns CONFLICT when the object does not exist, or has
                    # dynamodb attribute deleted=True. In either case, we treat
                    # the object as deleted
                    committed_to_appsync = True
                    object_exists = False
                elif e.errors[0].get("errorType") == "NOT_FOUND":
                    # Currently not returned, but should be instead of conflict
                    # when the object is deleted.
                    committed_to_appsync = True
                    object_exists = False
                else:
                    # Some other transient error
                    _rollback(obj)
                    db.session.rollback()
                    retries_remaining -= 1
            except Exception as e:
                current_app.logger.error(
                    "DELETE (hard) non-GraphQL exception {} {}: {}".format(
                        tag, obj.id, str(e)
                    )
                )
                _rollback(obj)
                db.session.rollback()
                raise errors.BadGatewayError()

    if committed_to_appsync:
        # If that was successful, try to delete the resources associated with
        # the object
        try:
            obj.delete(**kwargs)
        except Exception as e:
            current_app.logger.error("Object delete %s: %s", obj, str(e))
            # If that failed, restore the object to its previous state in appsync
            obj.deleted = obj_deleted
            current_app.logger.debug("Rolling back %s", obj)
            # Either of these could fail, leaving us in a state where the
            # object is marked as deleted in appsync, but still exists in
            # transfer/auth/dns. Since the model is marked as deleted in the
            # appsync representation, GCS will act as if it were deleted
            # as far as the API is concerned.
            if object_exists:
                current_app.logger.debug("Updating %s", obj)
                update_object(obj, _rollback)
            else:
                current_app.logger.debug("Recreating %s", obj)
                _rollback(obj)
                create_object(obj)
            if isinstance(e, errors.APIBaseError):
                raise
            else:
                raise errors.BadGatewayError(detail=str(e))
        # We succeeded in marking it as deleted and deleting it.
        if update:
            db.session.add(obj)
        else:
            db.session.delete(obj)
    else:
        # We couldn't talk to appsync, so the object is retained and we'll
        # return an error
        obj.deleted = obj_deleted
        db.session.rollback()
        raise errors.BadGatewayError()

    db.session.commit()


def validate_unique_custom_domain(domain_name: str) -> None:
    """
    Check that the domain name is both composed of a set of valid dns labels and
    that it is not in use by any currently existing collections or by the
    endpoint itself.

    :param domain_name:
        Domain name to check

    :raises errors.ValidationError:
        The domain_name is not a valid domain name string.
    :raises errors.ConflictError:
        The domain name is already in use by another object in this endpoint.
    """
    for label in domain_name.split("."):
        if not re.match(r"^(?![0-9]+$)(?!-)[a-zA-Z0-9-]{,63}(?<!-)$", label):
            raise errors.ValidationError("Invalid domain name value: {}".format(label))

    in_use = Collection.query.filter(
        Collection.domain_name == domain_name
    ).one_or_none()

    if in_use:
        raise errors.ConflictError("Domain already in use by a collection")

    ep = Endpoint.query.one()
    if ep.has_domain and ep.domain.domain_name == domain_name:
        raise errors.ConflictError("Domain already in use by the endpoint")
