import datetime
from typing import Any, Dict, Optional, Union
import uuid

from flask import current_app
import marshmallow as ma

from .configuration_object import ConfigurationObject
from globus.manager.api.schemas.appsync import AcmeAccount as acme_account_schema
from globus.manager.utils.acme_client import AcmeClient
from globus.manager.constants import ConfigurationObjectTag


class AcmeAccount(ConfigurationObject):
    __mapper_args__ = {"polymorphic_identity": ConfigurationObjectTag.ACME_ACCOUNT.name}
    epoch = datetime.datetime.fromtimestamp(0, datetime.timezone.utc)
    delta_zero = datetime.timedelta(seconds=0)
    _DATA_TYPE = "acme_account#1.0.0"

    configuration_schema = acme_account_schema

    _acme_client = None

    def __init__(self, appsync_data=None):
        if appsync_data is not None:
            args = {k: v for k, v in appsync_data.items()}
            args["tag"] = ConfigurationObjectTag.ACME_ACCOUNT
            args["id"] = ConfigurationObjectTag.ACME_ACCOUNT.name
            super().__init__(**args)
        else:
            super().__init__(
                tag=ConfigurationObjectTag.ACME_ACCOUNT,
                id=ConfigurationObjectTag.ACME_ACCOUNT.name,
            )
            self._acme_client = AcmeClient(
                current_app.dns_client,
                acme_key=None,
                acme_account=None,
            )
            self.acme_key = self._acme_client.acme_key
            self.acme_account = self._acme_client.acme_account
            self.configuration_hash["lease_node_id"] = None
            self.configuration_hash["lease_expiration"] = None

    def init_appsync(self):
        self._acme_client = None

    @property
    def acme_client(self):
        if self._acme_client is None:
            self._acme_client = AcmeClient(
                current_app.dns_client,
                acme_key=self.configuration_hash["acme_key"],
                acme_account=self.configuration_hash["acme_account"],
            )
        return self._acme_client

    @property
    def acme_account(self) -> Dict[str, Any]:
        value = self.configuration_hash.get("acme_account")
        if not value:
            raise AttributeError(
                "The acme_account property is not defined for this object"
            )
        return value

    @acme_account.setter
    def acme_account(self, value: Optional[Dict[str, Any]]) -> None:
        self.configuration_hash["acme_account"] = value
        if self._acme_client and self._acme_client.acme_account != value:
            self._acme_client = None

    @property
    def acme_key(self) -> Dict[str, Any]:
        value = self.configuration_hash.get("acme_key")
        if not value:
            raise AttributeError("The acme_key property is not defined for this object")

        return value

    @acme_key.setter
    def acme_key(self, value: Optional[Dict[str, Any]]):
        self.configuration_hash["acme_key"] = value
        if self._acme_client and self._acme_client.acme_key != value:
            self._acme_client = None

    @property
    def lease_node_id(self) -> str:
        value = self.configuration_hash.get("lease_node_id")

        if not value:
            raise AttributeError("lease_node_id")

        return value

    @lease_node_id.setter
    def lease_node_id(self, value: Union[str, uuid.UUID, None]):
        node_id = str(value) if value else None
        if self._lease_change_allowed(node_id):
            self.configuration_hash["lease_node_id"] = node_id
        else:
            raise ValueError(
                "Lease is owned by {} until {}".format(
                    self.lease_node_id, self.lease_expiration
                )
            )

    @property
    def lease_expiration(self) -> datetime.datetime:
        value = self.configuration_hash.get("lease_expiration")
        if value:
            return ma.utils.from_iso_datetime(value)

        raise AttributeError(
            "The lease_expiration property is not defined for this object"
        )

    @lease_expiration.setter
    def lease_expiration(self, value: Optional[datetime.datetime]):
        if value:
            self.configuration_hash["lease_expiration"] = ma.utils.isoformat(value)
        elif "lease_expiration" in self.configuration_hash:
            del self.configuration_hash["lease_expiration"]

    def acquire_lease(self, duration: Optional[datetime.timedelta] = None):
        self.lease_node_id = current_app.config["NODE_ID"]
        if duration is None:
            duration = datetime.timedelta(hours=12)

        utcnow = datetime.datetime.now(datetime.timezone.utc)

        self.lease_expiration = utcnow + duration

    def lease_remaining(self):
        utcnow = datetime.datetime.now(datetime.timezone.utc)
        try:
            lease_node_id = self.lease_node_id
        except AttributeError:
            lease_node_id = None

        if lease_node_id == current_app.config["NODE_ID"]:
            delta = self.lease_expiration - utcnow

            if delta > self.delta_zero:
                return delta
        return self.delta_zero

    def release_lease(self):
        if self.lease_node_id == current_app.config["NODE_ID"]:
            self.lease_expiration = self.epoch
            self.lease_node_id = None
        else:
            raise ValueError("This node does not have a current lease")

    def get_cert_chain_and_key(self, domain_name):
        current_lease_node_id = getattr(self, "lease_node_id", None)
        utcnow = datetime.datetime.now(datetime.timezone.utc)
        if (
            current_lease_node_id != current_app.config["NODE_ID"]
            or (self.lease_expiration or self.epoch) < utcnow
        ):
            raise ValueError("This node does not have a current lease")

        return self.acme_client.get_cert_chain_and_key(domain_name)

    def _lease_change_allowed(self, node_id=None):
        now = datetime.datetime.now(datetime.timezone.utc)
        if hasattr(self, "lease_node_id"):
            current_node = self.lease_node_id
        else:
            current_node = None

        if hasattr(self, "lease_expiration"):
            current_expiration = self.lease_expiration
        else:
            current_expiration = self.epoch
        # We want this to restrict changes to this so only one renew can be
        # in progress.
        #
        # Allow updates if
        #   -   current_expiration is unset or in the past
        #   -   the node matches the old node (either lease extension or
        #       termination)
        return (current_expiration is None or current_expiration < now) or (
            node_id == current_node
        )

    def rollback(self, data=None):
        """
        Rolls back state to a previous version of the configuration_hash.
        """
        if data:
            self.update(data["items"][0])

        self.init_appsync()
        self.refresh_configuration_hash()

    def __repr__(self):
        return "<AcmeAccount {}>".format(self.id)
