from flask import current_app

from .connector import Connector
from .decorators import checks_policy_property
from .storage_gateway import StorageGateway

from globus.manager.constants import ConfigurationObjectTag


class AzureBlobStorageGateway(StorageGateway):
    CONNECTOR_ID = "9436da0c-a444-11eb-af93-12704e0d6a4d"
    POLICY_DATA_TYPES = ["azure_blob_storage_policies#1.0.0"]
    POLICY_DATA_TYPE = "azure_blob_storage_policies#1.0.0"

    __mapper_args__ = {
        "polymorphic_identity": (
            ConfigurationObjectTag.STORAGE_GATEWAY.name + CONNECTOR_ID
        )
    }
    auth_types = [
        "user",
        "service_principal",
        "user_service_principal",
    ]
    connector_policy_names = [
        "client_id",
        "secret",
        "tenant",
        "account",
        "auth_type",
        "adls",
    ]

    connector = Connector(
        id=CONNECTOR_ID, display_name="Azure Blob", can_post_credential=False
    )

    auth_type = StorageGateway.connector_property("auth_type")

    @checks_policy_property(prop="auth_type")
    def check_auth_type(self, value: str):
        if value is not None and value not in self.auth_types:
            raise ValueError()

    client_id = StorageGateway.connector_property("client_id", required=True)

    @checks_policy_property(prop="client_id", required=True)
    def check_client_id(self, value: str):
        if not isinstance(value, str):
            raise ValueError()

    secret = StorageGateway.connector_property("secret", required=True)

    @checks_policy_property(prop="secret", required=True)
    def check_secret(self, value: str):
        if not isinstance(value, str):
            raise ValueError()

    tenant = StorageGateway.connector_property("tenant")

    @checks_policy_property(prop="tenant")
    def check_tenant(self, value: str):
        if value is not None and not isinstance(value, str):
            raise ValueError("Invalid tenant")

    account = StorageGateway.connector_property("account", required=True)

    @checks_policy_property(prop="account", required=True)
    def check_account(self, value: str):
        if not isinstance(value, str):
            raise ValueError()

    adls = StorageGateway.connector_property("adls")

    @checks_policy_property(prop="adls")
    def check_adls(self, value: bool):
        if not isinstance(value, bool):
            raise ValueError()

    def check_user_allowed(self, username: str):
        if self.client_id is None:
            current_app.logger.warning(
                "storage_gateway=%s client_id is not configured", self
            )
            return False

        if self.secret is None:
            current_app.logger.warning(
                "storage_gateway=%s secret is not configured", self
            )
            return False

        return super().check_user_allowed(username)

    def default_mapping(self, identity_data):
        if len(self.allowed_domains) == 1:
            user, domain = identity_data["username"].rsplit("@", 1)
            if domain in self.allowed_domains:
                return identity_data["username"]

    @property
    def user_credential_required(self):
        return getattr(self, "auth_type", None) != "service_principal"

    @property  # type: ignore
    def load_dsi_module(self) -> str:
        return "azure_blob"

    @load_dsi_module.setter
    def load_dsi_module(self, value: str):
        if value is not None and value != self.load_dsi_module:
            raise ValueError("Invalid load_dsi_module")

    def home_directory(self, username: str):
        return "/"

    def delete(self):
        self.client_id = None
        self.secret = None

        super().delete()

    @property
    def auth_callback(self):
        """
        URL of the auth callback that must be registered on the Microsoft API
        console for the application client_id in order to process Microsoft
        credentials.
        """
        try:
            return current_app.config["connectors"][self.CONNECTOR_ID].callback_url
        except Exception:
            raise AttributeError(
                "This storage gateway does not have a value for the auth_callback property"
            )

    @property  # type: ignore
    def process_user(self) -> str:
        return "gcsweb"

    @process_user.setter
    def process_user(self, value: str):
        if value is not None and value != self.process_user:
            raise ValueError("Invalid process_user")
