import pwd
from typing import Any, Dict, Optional, TYPE_CHECKING

from flask import current_app, g as flask_g

from globus.manager.api.models.connector import Connector
from globus.manager.constants import ConfigurationObjectTag
from .decorators import checks_policy_property
from .storage_gateway import StorageGateway

if TYPE_CHECKING:
    from globus.manager.api.connectors.hpss import HPSSConnector


class HPSSStorageGateway(StorageGateway):
    CONNECTOR_ID = "fb656a17-0f69-4e59-95ff-d0a62ca7bdf5"

    # Current policy data_type, used for new docs and dumping
    POLICY_DATA_TYPE = "hpss_storage_policies#1.0.0"
    # Supported policy data_types, current and legacy
    POLICY_DATA_TYPES = [POLICY_DATA_TYPE]

    __mapper_args__ = {
        "polymorphic_identity": (
            ConfigurationObjectTag.STORAGE_GATEWAY.name + CONNECTOR_ID
        )
    }

    # Our properties used by StorageGateway to validate input
    connector_policy_names = [
        "authentication_mech",
        "authenticator",
        "uda_checksum",
    ]

    connector = Connector(id=CONNECTOR_ID, display_name="HPSS")

    authentication_mech = StorageGateway.connector_property(
        "authentication_mech", required=True
    )

    @checks_policy_property(prop="authentication_mech", required=True)
    def _check_authentication_mech(self, value: str) -> None:
        if value not in ["unix", "krb5"]:
            raise ValueError()

    authenticator = StorageGateway.connector_property("authenticator", required=True)

    @checks_policy_property(prop="authenticator", required=True)
    def _check_authenticator(self, value) -> None:
        if not isinstance(value, (bytes, str)):
            raise ValueError()

    uda_checksum = StorageGateway.connector_property("uda_checksum", required=True)

    @checks_policy_property(prop="uda_checksum", required=True)
    def _check_uda_checksum(self, value: bool) -> None:
        if not isinstance(value, bool):
            raise ValueError()

    def default_mapping(self, identity_data: Dict[str, Any]) -> Optional[str]:
        if len(self.allowed_domains) == 1:
            user, domain = identity_data["username"].rsplit("@", 1)
            if domain in self.allowed_domains:
                return user
        return None

    def home_directory(self, username: str) -> Optional[str]:
        """
        Look up the user's home directory by calling a helper function built into
        the HPSS DSI. Used for expansion of '~' and  $HOME in path restrictions. The
        call into the DSI is handled by HPSSConnector.

        We cache the response and stuff it into flask.g so that it will persist
        between instances of HPSSStorageGateway. We make no attempts to lookup a user's
        home directory more than once; if the lookup fails the first time, we still
        return None for the lifetime of this request.
        """
        current_app.logger.debug(f"Looking up home directory for {username}")

        # Initialize the cache if it doesn't already exist
        if not hasattr(flask_g, "hpss_home_dir_cache"):
            flask_g.hpss_home_dir_cache = {}

        # Return any previous result, if one exists.
        if username in flask_g.hpss_home_dir_cache:
            home_dir = flask_g.hpss_home_dir_cache[username]
            current_app.logger.debug(f"Using cached home directory {home_dir}")
            return home_dir

        # Lookup the HPSSConnector instancec
        hpss_connector: HPSSConnector = current_app.config["connectors"][
            self.CONNECTOR_ID
        ]

        # Lookup the user's home directory and cache it
        flask_g.hpss_home_dir_cache[username] = hpss_connector.home_directory(
            username, self.authentication_mech, self.authenticator
        )

        # Return our (newly) cached response
        home_dir = flask_g.hpss_home_dir_cache[username]
        current_app.logger.debug(f"User's home directory is {home_dir}")
        return home_dir

    def lookup_unix_info(self, username: str) -> Optional[pwd.struct_passwd]:
        """
        Helper method to resolve the given username to a passwd entry.

        Although we rely on the HPSS API for the user's home directory,
        everything else can come from the system's accounts. This is
        possible because we have always required HPSS sites to have user
        accounts in /etc/passwd / nis / ldap for use with GridFTP server
        gfs_stat_t translations (back when we cared about that).
        """

        try:
            passwd_entry: Optional[pwd.struct_passwd] = pwd.getpwnam(username)
        except KeyError:
            passwd_entry = None

        if not passwd_entry:
            current_app.logger.info(
                "lookup_unix_info: no local account for user=%s", username
            )

        return passwd_entry

    def get_user_display_name(self, username: str) -> str:
        """
        Helper used to standardize display names for this connector.
        """
        pwd = self.lookup_unix_info(username)

        if pwd:
            return pwd.pw_gecos.split(",")[0] or username
        else:
            return username

    @property  # type: ignore
    def load_dsi_module(self) -> str:
        return "hpss"

    @load_dsi_module.setter
    def load_dsi_module(self, value: str) -> None:
        pass

    @property  # type: ignore
    def process_user(self) -> str:
        return "gcsweb"

    @process_user.setter
    def process_user(self, value: Any) -> None:
        pass

    def repr(self) -> str:
        return "<HPSSStorageGateway {}>".format(self.id)
