from typing import List, Optional
from globus.manager.constants import ConfigurationObjectTag

from .user_credential import UserCredential
from .decorators import checks_policy_property


class S3UserCredential(UserCredential):
    POLICY_DATA_TYPE = "s3_user_credential_policies#1.1.0"
    CONNECTOR_POLICIES = {
        "s3_user_credential_policies#1.0.0": [
            "s3_key_id",
            "s3_secret_key",
        ],
        "s3_user_credential_policies#1.1.0": [
            "s3_key_id",
            "s3_secret_key",
            "s3_requester_pays",
        ],
    }
    POLICY_DATA_TYPES = list(CONNECTOR_POLICIES.keys())
    CONNECTOR_ID = "7643e831-5f6c-4b47-a07f-8ee90f401d23"

    __mapper_args__ = {
        "polymorphic_identity": (
            ConfigurationObjectTag.USER_CREDENTIAL.name + CONNECTOR_ID
        )
    }

    anonymous_display_name = "Anonymous S3"
    unconfigured_display_name = "Unprovisioned S3 Credential"

    @property
    def connector_properties(self) -> List[str]:
        return ["s3_key_id", "s3_secret_key", "s3_requester_pays"]

    def __init__(self, appsync_data=None, doc=None, storage_gateway=None):
        if doc:
            invalid = False
            provisioned = False
            policies = doc.get("policies", {})
            s3_key_id = policies.get("s3_key_id")
            s3_secret_key = policies.get("s3_secret_key")
            s3_requester_pays = policies.get("s3_requester_pays")

            if storage_gateway.s3_user_credential_required:
                if s3_key_id:
                    display_name = s3_key_id
                else:
                    display_name = self.unconfigured_display_name
                if not policies:
                    policies = {
                        "DATA_TYPE": self.POLICY_DATA_TYPE,
                    }
                if not (s3_key_id and s3_secret_key):
                    invalid = True
                elif (
                    getattr(storage_gateway, "s3_requester_pays", False)
                    and not s3_requester_pays
                ):
                    raise AttributeError(
                        "The s3_requester_pays property is required for this credential."
                    )
                else:
                    provisioned = True
            else:
                display_name = self.anonymous_display_name
                if s3_key_id or s3_secret_key:
                    raise ValueError("key not allowed for anonymous access")
                if s3_requester_pays:
                    raise ValueError(
                        "s3_equester_pays not allowed for anonymous access"
                    )
                invalid = False
                provisioned = True
                policies = {
                    "DATA_TYPE": self.POLICY_DATA_TYPE,
                }
            doc = {
                **doc,
                **{
                    "display_name": display_name,
                    "invalid": invalid,
                    "deleted": False,
                    "provisioned": provisioned,
                    "policies": policies,
                },
            }
        super().__init__(
            appsync_data=appsync_data, doc=doc, storage_gateway=storage_gateway
        )

    @property
    def connector_id(self):
        return self.CONNECTOR_ID

    @connector_id.setter
    def connector_id(self, value):
        pass

    @property
    def s3_key_id(self) -> str:
        value = self.policies.get("s3_key_id")
        if value is None:
            raise AttributeError(
                "The s3_key_id property is not defined for this user credential"
            )
        return value

    @s3_key_id.setter
    def s3_key_id(self, value: Optional[str]) -> None:
        if self.storage_gateway.s3_user_credential_required:
            self.policies["s3_key_id"] = value
            if value is None:
                self.invalid = True
                self.display_name = "Unprovisioned S3 Credential"

    @s3_key_id.deleter
    def s3_key_id(self) -> None:
        self.policies["s3_key_id"] = None
        if self.storage_gateway.s3_user_credential_required:
            self.invalid = True
            self.display_name = "Unprovisioned S3 Credential"

    @checks_policy_property(prop="s3_key_id")
    def check_s3_key_id(self, value):
        if not self.storage_gateway.s3_user_credential_required and value:
            raise AttributeError()

    @property
    def s3_secret_key(self) -> str:
        value = self.policies.get("s3_secret_key")
        if value is None:
            raise AttributeError(
                "The s3_secret_key property is not defined for this user credential"
            )
        return value

    @s3_secret_key.setter
    def s3_secret_key(self, value):
        if self.storage_gateway.s3_user_credential_required:
            self.policies["s3_secret_key"] = value
            if value is None:
                self.invalid = True

    @s3_secret_key.deleter
    def s3_secret_key(self):
        self.policies["s3_secret_key"] = None
        if self.storage_gateway.s3_user_credential_required:
            self.invalid = True

    @checks_policy_property(prop="s3_secret_key")
    def check_s3_secret_key(self, value: str):
        if not self.storage_gateway.s3_user_credential_required and value:
            raise AttributeError()

    @property
    def s3_requester_pays(self) -> bool:
        value = self.policies.get("s3_requester_pays")
        if value is None:
            raise AttributeError(
                "The s3_requester_pays property is not defined for this user credential"
            )
        return value

    @s3_requester_pays.setter
    def s3_requester_pays(self, value):
        if self.storage_gateway.s3_user_credential_required:
            self.policies["s3_requester_pays"] = value
            if getattr(self.storage_gateway, "s3_requester_pays", False) and not value:
                self.invalid = True

    @s3_requester_pays.deleter
    def s3_requester_pays(self):
        self.policies["s3_requester_pays"] = None
        if getattr(self.storage_gateway, "s3_requester_pays", False):
            self.invalid = True

    @checks_policy_property(prop="s3_requester_pays")
    def check_s3_requester_pays(self, value: bool):
        if not self.storage_gateway.s3_user_credential_required and value:
            raise AttributeError()
        if getattr(self.storage_gateway, "s3_requester_pays", False) and not value:
            raise AttributeError(
                "The s3_requester_pays property is required for this credential."
            )

    def patch(self, **doc):
        doc.pop("display_name", None)
        policies = doc.get("policies", {})

        if self.storage_gateway.s3_user_credential_required:
            if policies is None:
                s3_key_id = None
                s3_secret_key = None
                s3_requester_pays = None
            else:
                s3_key_id = policies.get("s3_key_id", getattr(self, "s3_key_id", None))
                s3_secret_key = policies.get(
                    "s3_secret_key", getattr(self, "s3_secret_key", None)
                )
                s3_requester_pays = policies.get(
                    "s3_requester_pays", getattr(self, "s3_requester_pays", None)
                )

            if not s3_key_id or not s3_secret_key:
                if not doc.get("invalid", True):
                    raise ValueError(
                        "Cannot mark this credential as valid without keys"
                    )
                doc["invalid"] = True
            elif (
                getattr(self.storage_gateway, "s3_requester_pays", False)
                and not s3_requester_pays
            ):
                if not doc.get("invalid", True):
                    raise ValueError(
                        "Cannot mark this credential as valid without requester pays"
                    )
                doc["invalid"] = True
            else:
                doc["provisioned"] = True

            doc["display_name"] = self._provisioned_display_name(s3_key_id)

        super().patch(**doc)

    def put(self, **doc):
        invalid = doc.get("invalid")
        policies = doc.get("policies", {})
        s3_key_id = policies.get("s3_key_id")
        s3_secret_key = policies.get("s3_secret_key")
        s3_requester_pays = policies.get("s3_requester_pays")

        if self.storage_gateway.s3_user_credential_required:
            if not policies:
                policies = {
                    "DATA_TYPE": self.POLICY_DATA_TYPE,
                }
            if not (s3_key_id and s3_secret_key):
                if invalid is False:
                    raise ValueError("Cannot mark this credential as valid")
                invalid = True
            elif (
                getattr(self.storage_gateway, "s3_requester_pays", False)
                and not s3_requester_pays
            ):
                raise AttributeError(
                    "The s3_requester_pays property is required for this credential."
                )
            else:
                self.provisioned = True
        else:
            if s3_key_id or s3_secret_key:
                raise ValueError("key not allowed for anonymous access")
            if s3_requester_pays:
                raise ValueError("requester_pays not allowed for anonymous access")
            policies = {
                "DATA_TYPE": self.POLICY_DATA_TYPE,
            }
            self.provisioned = True
        if invalid is None:
            invalid = False

        doc["invalid"] = invalid
        doc["display_name"] = self._provisioned_display_name(s3_key_id)

        super().put(**doc)

    def _provisioned_display_name(self, s3_key_id):
        if self.storage_gateway.s3_user_credential_required:
            if s3_key_id:
                display_name = s3_key_id
            else:
                display_name = self.unconfigured_display_name
        else:
            display_name = self.anonymous_display_name

        return display_name

    def delete(self):
        self.s3_secret_key = None
        self.s3_requester_pays = None
        super().delete()

    def __repr__(self):
        return "<S3UserCredential {}>".format(self.id)
