from typing import Optional
import uuid

import marshmallow as ma
from typing_extensions import Literal, TypedDict

from .data_type import data_type, data_type_schema, Schema
from .schema_version_checker import verify_schema_version

_principal_type = {
    "urn:globus:auth:identity": "identity",
    "urn:globus:groups:id": "group",
}
_principal_re = r"^(urn:globus:auth:identity|urn:globus:groups:id):([a-f0-9-]*)$"


class RoleDict(TypedDict, total=False):
    DATA_TYPE: Literal["role#1.0.0"]
    id: uuid.UUID
    principal: str
    collection: Optional[str]
    role: str


@data_type("role#1.0.0")
class Role_1_0_0(Schema):
    """
    The "Role" document type represents the assignment of a role on an
    Endpoint or Collection to a Globus identity or group.
    """

    id = ma.fields.UUID(
        description="""
        Unique id string for this role assignment. This is system generated
        and should not be included in create requests.
        """
    )
    resource = ma.fields.UUID(
        data_key="collection",
        required=False,
        allow_none=True,
        description="""
        Collection Id. This value is omitted when creating an endpoint role
        or when creating role definitions when creating "collections.
        """,
    )
    principal = ma.fields.String(
        required=True,
        validate=[ma.validate.Regexp(_principal_re)],
        description="Globus Auth identity or group id URN",
    )
    role_name = ma.fields.String(
        data_key="role",
        required=True,
        validate=[
            ma.validate.OneOf(
                [
                    "owner",
                    "administrator",
                    "access_manager",
                    "activity_manager",
                    "activity_monitor",
                    "access_monitor",
                ]
            )
        ],
        description="Role assigned to the principal",
    )

    # Marshmallow claims that "attribute" can be used to serialize a particular
    # attribute name, but it also attempts to *deserialize* that name, so it
    # doesn't work for our case where model property doesn't match the API schema.
    # Similarly, using data_key works for serialization, but on
    # deserialization, it changes the name to match property instead of the data key.
    # This rewrites the schema properties with data_keys to match those in the values
    @ma.post_load
    def rewrite_data_keys(self, data, **kwargs):
        data["role"] = data.pop("role_name")
        if "resource" in data:
            data["collection"] = data.pop("resource")
        return data

    @ma.validates("DATA_TYPE")
    def validate_data_type(self, data, **kwargs):
        verify_schema_version(data, self._DATA_TYPE)

    def load(self, *args, **kwargs) -> RoleDict:
        return super().load(*args, **kwargs)


Role = data_type_schema.Role
