import typing as t
import uuid

import marshmallow as ma

from .data_type import data_type, data_type_schema, PolicySchema, Schema
from .identity_mapping import IdentityMapping
from .path_restrictions import PathRestrictions
from .schema_version_checker import version_safe_type_schemas


class StorageGatewayPolicies(ma.fields.Method):
    connector_policies: t.ClassVar = version_safe_type_schemas[str, PolicySchema]()
    data_types: t.ClassVar[t.Dict[str, PolicySchema]] = {}

    def _jsonschema_type_mapping(self) -> t.Dict[str, t.Any]:
        type_mapping: t.Dict[str, t.Any] = {
            "oneOf": [],
        }
        refs = type_mapping["oneOf"]
        for policy in self.__class__.data_types.values():
            title = getattr(
                policy, "title", policy.__class__.__name__.replace("Schema", "")
            )
            refs.append({"$ref": f"#/components/schemas/{title}"})

        metadata = self.metadata.get("metadata", {})
        metadata.update(self.metadata)
        for md_key, md_val in metadata.items():
            if md_key in ("metadata", "name"):
                continue
            type_mapping[md_key] = md_val
        return type_mapping

    @classmethod
    def register_schema(
        cls, schema: t.Type[PolicySchema], register_data_type: bool = True
    ) -> None:
        cls.connector_policies[schema.connector_id] = schema.public_schema
        if register_data_type:
            cls.data_types[schema._DATA_TYPE] = schema.public_schema


T = t.TypeVar("T", bound=t.Type[PolicySchema])


def storage_gateway_policy(
    discriminator: str,
    register_data_type: bool = True,
    schema_base_name: t.Optional[str] = None,
) -> t.Callable[[T], T]:
    def decorator(cls: T) -> T:
        decorated_cls = t.cast(
            T,
            data_type(
                discriminator,
                register_data_type=register_data_type,
                schema_base_name=schema_base_name,
            )(cls),
        )

        StorageGatewayPolicies.register_schema(
            decorated_cls, register_data_type=register_data_type
        )

        return decorated_cls

    return decorator


@data_type("storage_gateway#1.0.0")
class StorageGateway_1_0_0(Schema):
    __doc__ = """
    A storage gateway provides the access policies for the endpoint's
    connected storage systems. It is a named interface by which
    authorized users can create and manage collections on the
    connected storage system. A single storage system may be
    associated with multiple storage gateways, each with its own
    policies.

    Storage gateway policies describe what type connector the storage
    gateway uses, the paths it allows access to, the login
    requirements are for the storage gateway, and the algorithm to
    map Globus identities to the user namespace of the storage
    gateway (e.g. local accounts).
    """

    id = ma.fields.UUID(description="Unique id for this storage gateway")
    display_name = ma.fields.String(
        validate=[ma.validate.Length(max=256)],
        description="Name of the storage gateway",
    )
    connector_id = ma.fields.UUID(
        description="""
        Id of the connector type that this storage gateway interacts with.
        """
    )
    high_assurance = ma.fields.Boolean(
        required=False,
        allow_none=True,
        description="""
        Flag indicating if the storage_gateway requires high
        assurance features.
        """,
    )
    require_high_assurance = ma.fields.Boolean(
        required=False,
        allow_none=True,
        description="Alias for **high_assurance**",
        deprecated=True,
    )
    authentication_assurance_timeout = ma.fields.Int(
        required=False,
        allow_none=True,
        validate=[ma.validate.Range(min=1)],
        load_only=True,
        description="Alias for **authentication_timeout_mins**",
        deprecated=True,
    )
    authentication_timeout_mins = ma.fields.Int(
        required=False,
        allow_none=True,
        validate=[ma.validate.Range(min=1)],
        description="""
        Timeout (in minutes) during which a user is required to have
        authenticated to access files or create user credentials on this
        storage gateway.

        For a high assurance storage gateway, this must be done within the
        current Globus Auth session, otherwise, the caller can perform the
        authentication with any application which uses Globus Auth.
        """,
    )
    allowed_domains = ma.fields.List(
        ma.fields.String,
        description="""
        List of allowed domains. Users creating credentials or collections
        on this storage gateway must have an identity in one of these domains.
        """,
    )
    identity_mappings = ma.fields.List(
        ma.fields.Nested(IdentityMapping),
        required=False,
        allow_none=True,
        description="""
        List of identity mappings to apply to user identities to determine
        what connector-specific accounts are available for access.
        """,
    )
    users_allow = ma.fields.List(
        ma.fields.Str,
        required=False,
        allow_none=True,
        description="""
        List of connector-specific usernames allowed to access this storage
        gateway.
        """,
    )
    users_deny = ma.fields.List(
        ma.fields.Str,
        required=False,
        allow_none=True,
        description="""
        List of connector-specific usernames denied access to this storage
        gateway.
        """,
    )
    restrict_paths = ma.fields.Nested(
        PathRestrictions,
        required=False,
        allow_none=True,
        description="""
        Path restrictions within this storage gateway. Paths are
        interpreted as absolute paths in the file namespace of the
        connector.
        """,
    )
    policies = StorageGatewayPolicies(
        serialize="dump_policies",
        deserialize="load_policies",
        description="Connector-specific storage policies",
    )
    process_user = ma.fields.Str(
        required=False,
        allow_none=True,
        description="""
        Local POSIX user the GridFTP server should run as when accessing
        this storage gateway.
        """,
    )
    load_dsi_module = ma.fields.Str(
        required=False,
        allow_none=True,
        description="""
        Name of the DSI module to load by the GridFTP server when
        accessing this storage gateway.
        """,
    )
    deleted = ma.fields.Boolean(
        required=False,
        description="Flag indicating that this storage gateway has been deleted",
    )

    def load_policies(self, value) -> t.Dict[str, t.Any]:
        data_type = value["DATA_TYPE"]

        if data_type not in StorageGatewayPolicies.data_types:
            raise ma.ValidationError(
                f"Unsupported policies document DATA_TYPE {data_type}",
                field_name="policies",
            )
        if self.partial:
            schema = StorageGatewayPolicies.data_types[data_type].patch_schema
        else:
            schema = StorageGatewayPolicies.data_types[data_type]

        data = schema.load(value, partial=self.partial)
        return data

    def dump_policies(self, obj) -> t.Dict[str, t.Any]:
        connector_id = str(ma.utils.get_value(obj, "connector_id"))
        if connector_id not in StorageGatewayPolicies.connector_policies:
            raise ValueError("Unable to serialize {} {}".format(obj, connector_id))

        if not self.load_only:
            schema = StorageGatewayPolicies.connector_policies[
                connector_id
            ].private_schema
        else:
            schema = StorageGatewayPolicies.connector_policies[
                connector_id
            ].public_schema

        return schema.dump(obj)

    @ma.post_load
    def backward_compatible_assurance_timeout_load(self, obj, **kwargs):
        authentication_assurance_timeout = obj.pop(
            "authentication_assurance_timeout", None
        )
        if authentication_assurance_timeout:
            obj["authentication_timeout_mins"] = authentication_assurance_timeout
        return obj

    @ma.post_dump
    def backward_compatible_require_high_assurance_dump(self, data, **kwargs):
        data["require_high_assurance"] = data.get("high_assurance")
        return data

    @ma.post_load
    def backward_compatible_require_high_assurance_load(self, obj, **kwargs):
        if "high_assurance" not in obj and "require_high_assurance" in obj:
            obj["high_assurance"] = obj.pop("require_high_assurance")
        return obj

    @ma.post_load
    def stringify_uuid(self, obj, **kwargs):
        res = {}
        for k, v in obj.items():
            if isinstance(v, uuid.UUID):
                res[k] = str(v)
            elif isinstance(v, list):
                newlist = []
                for i in v:
                    if isinstance(i, uuid.UUID):
                        newlist.append(str(i))
                    else:
                        newlist.append(i)
                res[k] = newlist
            else:
                res[k] = v
        return res

    private = [
        "deleted",
        "identity_mappings",
        "restrict_paths",
        "process_user",
        "load_dsi_module",
        "users_allow",
        "users_deny",
        "authentication_assurance_timeout",  # private since it is deprecated
    ]


@data_type("storage_gateway#1.1.0")
class StorageGateway_1_1_0(StorageGateway_1_0_0):
    __doc__ = (
        StorageGateway_1_0_0.__doc__
        + """
    Version 1.1.0 includes support for multi-factor authentication requirements
    for high assurance storage gateways.
    """
    )

    require_mfa = ma.fields.Boolean(
        required=False,
        default=False,
        description="""
        Flag indicating if the storage_gateway requires multi-factor
        authentication. Only usable on high assurance storage gateways.
        """,
    )


@data_type("storage_gateway#1.2.0")
class StorageGateway_1_2_0(StorageGateway_1_1_0):
    __doc__ = (
        StorageGateway_1_1_0.__doc__
        + """
    Version 1.2.0 includes support for admin managed credentials.
    """
    )

    admin_managed_credentials = ma.fields.Boolean(
        required=False,
        default=False,
        description="""
        Flag indicating if the storage_gateway allows endpoint
        administrators to manage user credentials on behalf of other users.
        """,
    )


StorageGateway = data_type_schema.StorageGateway
