import typing as t

import marshmallow as ma

from .data_type import data_type, data_type_schema, PolicySchema, Schema
from .schema_version_checker import verify_schema_version, version_safe_type_schemas


class UserCredentialPolicies(ma.fields.Method):
    connector_policies: t.ClassVar = version_safe_type_schemas[str, PolicySchema]()
    data_types: t.ClassVar[t.Dict[str, PolicySchema]] = {}

    def _jsonschema_type_mapping(self):
        type_mapping = {
            "oneOf": [],
        }
        refs = type_mapping["oneOf"]
        for policy in self.__class__.data_types.values():
            title = getattr(
                policy, "title", policy.__class__.__name__.replace("Schema", "")
            )
            refs.append({"$ref": f"#/components/schemas/{title}"})

        metadata = self.metadata.get("metadata", {})
        metadata.update(self.metadata)
        for md_key, md_val in metadata.items():
            if md_key in ("metadata", "name"):
                continue
            type_mapping[md_key] = md_val
        return type_mapping

    @classmethod
    def register_schema(
        cls, schema: t.Type[PolicySchema], register_data_type: bool = True
    ) -> None:
        cls.connector_policies[schema.connector_id] = schema.public_schema
        if register_data_type:
            cls.data_types[schema._DATA_TYPE] = schema.public_schema


T = t.TypeVar("T", bound=t.Type[PolicySchema])


def user_credential_policy(
    discriminator: str,
    register_data_type: bool = True,
    schema_base_name: t.Optional[str] = None,
) -> t.Callable[[T], T]:
    def decorator(cls: T) -> T:
        decorated_cls = t.cast(
            T,
            data_type(
                discriminator,
                register_data_type=register_data_type,
                schema_base_name=schema_base_name,
            )(cls),
        )

        UserCredentialPolicies.register_schema(
            decorated_cls, register_data_type=register_data_type
        )

        return decorated_cls

    return decorator


@data_type("user_credential#1.0.0")
class UserCredential_1_0_0(Schema):
    """
    Credential information for an identity on a particular storage gateway and
    its related collections.
    """

    id = ma.fields.UUID(
        required=False, description="Unique id for this user credential"
    )
    identity_id = ma.fields.UUID(
        description="Globus Auth identity id that this credential is associated with"
    )
    connector_id = ma.fields.UUID(
        description="Id of the connector type used by this credential"
    )
    storage_gateway_id = ma.fields.UUID(
        description="Storage Gateway this credential is associated with"
    )
    username = ma.fields.Str(
        description="""
        Connector-specific username that this credential is associated
        with. If the connector supports identity mapping, this matches the
        result of the mapping applied to identity_id.
        """
    )
    display_name = ma.fields.Str(
        required=False,
        allow_none=True,
        validate=[ma.validate.Length(max=256)],
        description="Display name of the credential",
    )
    home_directory = ma.fields.String(
        dump_only=True,
        description="The home directory of this account associated with this credential",
    )
    invalid = ma.fields.Boolean(
        required=False,
        description="Flag indicating that this credential is no longer valid",
    )
    provisioned = ma.fields.Boolean(
        required=False,
        description="""
        Flag indicating that this credential has been fully provisioned. If
        this is false and the invalid property is true, then the credential
        was created during login and patching it to add the missing data
        should be presented to the user as initializing the credential.
        """,
    )
    deleted = ma.fields.Boolean(
        required=False,
        description="Flag indicating that this credential has been deleted",
    )
    policies = UserCredentialPolicies(
        serialize="dump_policies",
        deserialize="load_policies",
        description="Connector-specific user credential policies",
    )

    private = ["home_directory"]

    @ma.validates("DATA_TYPE")
    def validate_data_type(self, data, **kwargs):
        verify_schema_version(data, self._DATA_TYPE)

    @ma.pre_load
    def ignore_loading_home_directory(self, data, **kwargs):
        """
        Remove the home_directory property when loading. We want to serialize
        this when replying to gridftp login but don't want to care about a
        remotely loaded value.
        """
        data.pop("home_directory", None)
        return data

    def load_policies(self, value):
        data_type = value["DATA_TYPE"]

        if data_type not in UserCredentialPolicies.data_types:
            raise ma.ValidationError(
                f"Unsupported policies document DATA_TYPE {data_type}",
                field_name="policies",
            )
        if self.partial:
            schema = UserCredentialPolicies.data_types[data_type].patch_schema
        else:
            schema = UserCredentialPolicies.data_types[data_type]

        data = schema.load(value, partial=self.partial)
        return data

    def dump_policies(self, obj):
        connector_id = str(ma.utils.get_value(obj, "connector_id"))
        if connector_id not in UserCredentialPolicies.connector_policies:
            raise ValueError("Unable to serialize {} {}".format(obj, connector_id))

        if not self.load_only:
            schema = UserCredentialPolicies.connector_policies[
                connector_id
            ].private_schema
        else:
            schema = UserCredentialPolicies.connector_policies[
                connector_id
            ].public_schema

        return schema.dump(obj)

    @ma.post_dump
    def remove_deleted(self, data, **kwargs):
        if not data.get("deleted"):
            data.pop("deleted", None)
        return data


UserCredential = data_type_schema.UserCredential
