import marshmallow as ma

from ..api import (
    PolicySchema,
    collection_policy,
    data_type_schema,
    storage_gateway_policy,
    user_credential_policy,
)


AZURE_BLOB_CONNECTOR_ID = "9436da0c-a444-11eb-af93-12704e0d6a4d"


@storage_gateway_policy("azure_blob_storage_policies#1.0.0")
class AzureBlobStoragePolicies_1_0_0(PolicySchema):
    """
    Connector-specific storage gateway policies for the AzureBlob connector
    """

    auth_type = ma.fields.Str(
        required=True,
        allow_none=True,
        description="""
        The method of authentication to Azure. "user" prompts the user to
        log in to their Microsoft account via an oauth2 flow.
        "service_principal" uses the configured client_id and client_secret
        values to authenticate as an Azure service principal.
        """,
    )
    client_id = ma.fields.Str(
        required=True,
        allow_none=True,
        description="""
        Client ID registered with the Azure console to access Azure Blob.
        """,
    )
    secret = ma.fields.Str(
        required=True,
        allow_none=True,
        description="""
        Secret created in the Azure console to access Azure Blob with the
        client_id in this policy.
        """,
    )

    tenant = ma.fields.Str(
        required=True,
        allow_none=True,
        description="Tenant id of the Microsoft organization",
    )
    account = ma.fields.Str(
        required=True,
        allow_none=True,
        description="Azure Storage account to access with this storage gateway",
    )
    adls = ma.fields.Boolean(
        required=True,
        allow_none=True,
        description="""
        Flag indicating the Azure storage account has enabled Azure Data
        Lake Gen2 hierarchical namespace support.
        """,
    )
    user_credential_required = ma.fields.Bool(
        dump_only=True,
        description="""
            Flag indicating whether users must register a credential.  If true
            (or if this property is missing), this storage gateway is
            configured for OAuth2 user authentication.  If false,
            authentication is configured by the admin.
            """,
    )
    auth_callback = ma.fields.URL(
        schemes=["https"],
        dump_only=True,
        description="""
        URL of the auth callback that must be registered on the Microsoft
        API console for the application client_id in order to process
        Microsoft credentials.
        """,
    )

    private = ["client_id", "secret", "tenant", "account", "adls"]
    connector_id = AZURE_BLOB_CONNECTOR_ID

    @ma.pre_load
    def ignore_dump_only(self, data, **kwargs):
        """
        Ignore dump_only schema fields when loading.
        """
        for k, v in self.fields.items():
            if v.dump_only:
                data.pop(k, None)
        return data


@collection_policy("azure_blob_collection_policies#1.0.0")
class AzureBlobCollectionPolicies_1_0_0(PolicySchema):
    """
    Connector-specific collection policies for the AzureBlob connector
    """

    connector_id = AZURE_BLOB_CONNECTOR_ID


@user_credential_policy("azure_blob_user_credential_policies#1.0.0")
class AzureBlobUserCredentialPolicies_1_0_0(PolicySchema):
    """
    Connector-specific user credential policies for the AzureBlob connector
    """

    sub = ma.fields.Str(allow_none=True, description="OAuth subject identifier claim")
    email = ma.fields.Email(allow_none=True, description="OAuth email claim")
    access_token = ma.fields.Str(
        required=False, allow_none=True, description="OAuth access token"
    )
    refresh_token = ma.fields.Str(
        required=False,
        allow_none=True,
        description="OAuth refresh_token token",
    )
    token_expiry = ma.fields.AwareDateTime(
        description="OAuth access token expiration time"
    )
    scopes = ma.fields.List(
        ma.fields.Str,
        description="OAuth scopes associated with this access token",
    )
    tid = ma.fields.Str(description="Tenant id")

    private = ["access_token", "refresh_token"]
    connector_id = AZURE_BLOB_CONNECTOR_ID


AzureBlobCollectionPolicies = data_type_schema.AzureBlobCollectionPolicies
AzureBlobStoragePolicies = data_type_schema.AzureBlobStoragePolicies
AzureBlobUserCredentialPolicies = data_type_schema.AzureBlobUserCredentialPolicies
