import marshmallow as ma

from ..api import (
    PolicySchema,
    collection_policy,
    data_type_schema,
    storage_gateway_policy,
    user_credential_policy,
)


GOOGLE_CLOUD_STORAGE_CONNECTOR_ID = "56366b96-ac98-11e9-abac-9cb6d0d9fd63"


class GoogleCloudStorageProject(ma.Schema):
    """
    A Google Cloud Platform project resource
    """

    projectId = ma.fields.Str(
        description="Google-issued id of a Google Cloud Platform project"
    )
    name = ma.fields.Str(description="The name of the project")


@storage_gateway_policy("google_cloud_storage_policies#1.0.0")
class GoogleCloudStoragePolicies_1_0_0(PolicySchema):
    """
    Connector-specific storage gateway policies for the Google Cloud Storage
    connector
    """

    client_id = ma.fields.Str(
        required=True,
        allow_none=True,
        description="""
        Client ID registered with the Google Application console to access
        Google Cloud Storage.
        """,
    )
    secret = ma.fields.Str(
        required=True,
        allow_none=True,
        description="""
        Secret created to access access Google Cloud Storage with the client_id
        in this policy.
        """,
    )
    service_account_key = ma.fields.Raw(
        required=False,
        allow_none=True,
        description="Service account key to use when authenticating all storage access",
    )
    buckets = ma.fields.List(
        ma.fields.String(),
        required=False,
        allow_none=True,
        description="""
        The list of Google Cloud Storage buckets which the Storage Gateway is
        allowed to access, as well as the list of buckets that will be shown in
        root level directory listings. If this list is unset, bucket access is
        unrestricted and all non public credential accessible buckets will be
        shown in root level directory listings. The value is a list of bucket
        names.
        """,
    )
    projects = ma.fields.List(
        ma.fields.String(),
        required=False,
        allow_none=True,
        description="""
        The list of Google Cloud Storage project ids which the Storage Gateway
        is allowed to access. If this list is unset, project access is
        unrestricted. The value is a list of project id strings.
        """,
    )
    user_credential_required = ma.fields.Bool(
        dump_only=True,
        description="""
        Flag indicating whether users must register a credential.  If true (or
        if this property is missing), this storage gateway is configured for
        OAuth2 user authentication.  If false, authentication is configured by
        the admin.
        """,
    )
    auth_callback = ma.fields.URL(
        schemes=["https"],
        dump_only=True,
        description="""
        URL of the auth callback that must be registered on the Google API
        console for the application client_id in order to process "
        Google credentials.
        """,
    )

    private = ["client_id", "secret", "service_account_key"]
    connector_id = GOOGLE_CLOUD_STORAGE_CONNECTOR_ID

    @ma.pre_load
    def ignore_dump_only(self, data, **kwargs):
        """
        Ignore dump_only schema fields when loading.
        """
        for k, v in self.fields.items():
            if v.dump_only:
                data.pop(k, None)
        return data


@collection_policy("google_cloud_storage_collection_policies#1.0.0")
class GoogleCloudStorageCollectionPolicies_1_0_0(PolicySchema):
    """
    Connector-specific collection policies for the Google Cloud Storage
    connector
    """

    project = ma.fields.String(
        description="""
        Google Cloud Platform project ID value that is associated with this
        collection.
        """
    )
    connector_id = GOOGLE_CLOUD_STORAGE_CONNECTOR_ID


@user_credential_policy("google_cloud_storage_user_credential_policies#1.0.0")
class GoogleCloudStorageUserCredentialPolicies_1_0_0(PolicySchema):
    """
    Connector-specific user credential policies for the Google Cloud Storage
    connector
    """

    sub = ma.fields.Str(
        description="OpenID Connect subject property of this credential"
    )
    email = ma.fields.Email(
        description="OpenID Connect email property of this credential"
    )
    access_token = ma.fields.Str(
        required=False,
        allow_none=True,
        description="Access token to interact with the Google Cloud Storage API",
    )

    refresh_token = ma.fields.Str(
        required=False,
        allow_none=True,
        description="""
        Refresh token to generate new access tokens to use with the Google
        Cloud Storage API
        """,
    )
    token_expiry = ma.fields.AwareDateTime(
        description="Time when he access token expires"
    )
    scopes = ma.fields.List(
        ma.fields.Str,
        description="List of OAuth2 scopes associated with the tokens in this credential",
    )
    projects = ma.fields.List(
        ma.fields.Nested(GoogleCloudStorageProject),
        description="""
        List of Google Cloud Platform projects available for use with this
        credential.
        """,
    )

    private = ["access_token", "refresh_token"]
    connector_id = GOOGLE_CLOUD_STORAGE_CONNECTOR_ID


GoogleCloudStorageCollectionPolicies = (
    data_type_schema.GoogleCloudStorageCollectionPolicies
)
GoogleCloudStoragePolicies = data_type_schema.GoogleCloudStoragePolicies
GoogleCloudStorageUserCredentialPolicies = (
    data_type_schema.GoogleCloudStorageUserCredentialPolicies
)
