import marshmallow as ma

from ..api import (
    PolicySchema,
    collection_policy,
    data_type_schema,
    storage_gateway_policy,
    user_credential_policy,
)


HPSS_CONNECTOR_ID = "fb656a17-0f69-4e59-95ff-d0a62ca7bdf5"


@storage_gateway_policy(
    "hpss_storage_policies#1.0.0", schema_base_name="HPSSStoragePolicies"
)
class HPSSStoragePolicies_1_0_0(PolicySchema):
    """
    Connector-specific storage gateway policies for the HPSS connector
    """

    authentication_mech = ma.fields.String(
        validate=[ma.validate.OneOf(["krb5", "unix"])],
        required=True,
        description="""
        The type of authentication the connector will perform when logging into HPSS
        """,
    )
    authenticator = ma.fields.Str(
        required=True,
        validate=ma.validate.Regexp(r"^(auth_keyfile|auth_keytab):/"),
        description="""
        Authenticator used with authentication mech to perform authentication
        to HPSS. Format is: "<auth_type>:<auth_file>" where <auth_type> is one
        of "auth_keytab" or "auth_keyfile".
        """,
        example="auth_keytab:/var/hpss/etc/gridftp.keytab",
    )
    uda_checksum = ma.fields.Boolean(
        required=True,
        description="""
        Flag that indicates if checksums should be stored within UDAs so that
        sync-by-checksum transfers can verify the file without staging the file
        from tape.
        """,
    )

    private = ["authentication_mech", "authenticator"]
    connector_id = HPSS_CONNECTOR_ID


@collection_policy(
    "hpss_collection_policies#1.0.0", schema_base_name="HPSSCollectionPolicies"
)
class HPSSCollectionPolicies_1_0_0(PolicySchema):
    """
    Connector-specific collection policies for the HPSS connector
    """

    connector_id = HPSS_CONNECTOR_ID


@user_credential_policy(
    "hpss_user_credential_policies#1.0.0", schema_base_name="HPSSUserCredentialPolicies"
)
class HPSSUserCredentialPolicies_1_0_0(PolicySchema):
    """
    Connector-specific user credential policies for the HPSS connector
    """

    connector_id = HPSS_CONNECTOR_ID


HPSSCollectionPolicies = data_type_schema.HPSSCollectionPolicies
HPSSStoragePolicies = data_type_schema.HPSSStoragePolicies
HPSSUserCredentialPolicies = data_type_schema.HPSSUserCredentialPolicies
