import marshmallow as ma

from ..api import (
    PolicySchema,
    collection_policy,
    data_type_schema,
    storage_gateway_policy,
    user_credential_policy,
)


ONEDRIVE_CONNECTOR_ID = "28ef55da-1f97-11eb-bdfd-12704e0d6a4d"


@storage_gateway_policy(
    "onedrive_storage_policies#1.0.0", schema_base_name="OneDriveStoragePolicies"
)
class OneDriveStoragePolicies_1_0_0(PolicySchema):
    """
    Connector-specific storage gateway policies for the OneDrive connector
    """

    client_id = ma.fields.Str(
        required=True,
        allow_none=True,
        description="Client ID registered with the Azure console to access OneDrive",
    )
    secret = ma.fields.Str(
        required=True,
        allow_none=True,
        description="""
        Secret created in the Azure console to access OneDrive with the
        client_id in this policy.
        """,
    )
    tenant = ma.fields.Str(
        required=False,
        allow_none=True,
        description="""
        Tenant ID of the Microsoft organization. Required when Supported
        Account Types of the Azure application is set to Single tenant.
        """,
    )
    user_api_rate_limit = ma.fields.Int(
        required=False,
        description="""
        User API Rate Limit associated with this client ID in operations per
        second per user.
        """,
    )
    auth_callback = ma.fields.URL(
        schemes=["https"],
        dump_only=True,
        description="""
        URL of the auth callback that must be registered on the Microsoft
        API console for the application client_id in order to process
        Microsoft credentials.
        """,
    )

    private = ["client_id", "secret", "tenant", "user_api_rate_limit"]
    connector_id = ONEDRIVE_CONNECTOR_ID

    @ma.pre_load
    def ignore_dump_only(self, data, **kwargs):
        """
        Ignore dump_only schema fields when loading.
        """
        for k, v in self.fields.items():
            if v.dump_only:
                data.pop(k, None)
        return data


@collection_policy(
    "onedrive_collection_policies#1.0.0", schema_base_name="OneDriveCollectionPolicies"
)
class OneDriveCollectionPolicies_1_0_0(PolicySchema):
    """
    Connector-specific collection policies for the OneDrive connector
    """

    connector_id = ONEDRIVE_CONNECTOR_ID


@user_credential_policy(
    "onedrive_user_credential_policies#1.0.0",
    schema_base_name="OneDriveUserCredentialPolicies",
)
class OneDriveUserCredentialPolicies_1_0_0(PolicySchema):
    """
    Connector-specific user credential policies for the OneDrive connector
    """

    sub = ma.fields.Str(allow_none=True, description="OAuth subject identifier claim")
    email = ma.fields.Email(allow_none=True, description="OAuth email claim")
    access_token = ma.fields.Str(
        required=False,
        allow_none=True,
        description="OAuth access token",
    )

    refresh_token = ma.fields.Str(
        required=False,
        allow_none=True,
        description="OAuth refresh token",
    )
    token_expiry = ma.fields.AwareDateTime(
        description="OAuth access token expiration time"
    )
    scopes = ma.fields.List(
        ma.fields.Str,
        description="OAuth scopes associated with the access token",
        example=["openid", "email", "profile", "offline_access", "files.readwrite.all"],
    )

    tid = ma.fields.Str()

    private = ["access_token", "refresh_token"]
    connector_id = ONEDRIVE_CONNECTOR_ID


OneDriveCollectionPolicies = data_type_schema.OneDriveCollectionPolicies
OneDriveStoragePolicies = data_type_schema.OneDriveStoragePolicies
OneDriveUserCredentialPolicies = data_type_schema.OneDriveUserCredentialPolicies
