"""
globus-connect-server node create

The globus-connect-server node create command creates a node UUID and adds
a node key to the GCS endpoint's keychain. It outputs a node configuration
that can be used as an argument to the --import-node command-line option of
globus-connect-server node setup
"""
import json

import click

from gcs_cli.parsing.command_decorator import globus_command

from globus.manager.utils.keychain import NodeKey


class NodeKeyFileParam(click.File):
    name = "node_key_file"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def convert(self, value, param, ctx):
        with super().convert(value, param, ctx) as f:
            try:
                node_key = NodeKey(**json.load(f))
            except Exception as e:
                self.fail("Error parsing {}: {}".format(value, str(e)), param, ctx)
        return node_key


@globus_command(
    "create",
    short_help="Create a Globus Connect Server node configuration",
    help="Create a Globus Connect Server node configuration",
    disable_options=["format", "explicit_host"],
)
@click.option(
    "--client-id",
    "-c",
    help="Auth Client ID for this Globus Connect Server Endpoint",
    required=True,
)
@click.option(
    "--secret",
    "-s",
    help="Client secret for the auth client",
    prompt=True,
    hide_input=True,
    envvar="GLOBUS_CLIENT_SECRET",
    show_envvar=True,
)
@click.option(
    "--deployment-key",
    help="Path to deployment key",
    default="deployment-key.json",
    type=NodeKeyFileParam(),
)
@click.option(
    "--export-node",
    type=click.Path(dir_okay=False, writable=True),
    required=True,
    metavar="NODE_INFO_JSON",
    help="File to write node configuration to for deployment.",
)
def node_create(client_id, secret, deployment_key, export_node):
    from .node_create import node_create as nc

    nc(client_id, secret, deployment_key, export_node)
