import json

from jwcrypto import jwe, jwk
from flask import current_app

from globus.manager.constants import (
    ConfigurationObjectTag,
    DATA_ALG,
    DATA_ENC,
    DATA_KEY_SIZE,
    NODE_ALG,
    NODE_ENC,
    NODE_KEY_SIZE,
)
from globus.manager.api.errors import LimitExceededError

from globus.manager.api.schemas.appsync.keychain import KeyChain as keychain_schema

keychain_schema = keychain_schema.private_schema


class DataKeyError(Exception):
    pass


class NodeKey(jwk.JWK):
    @staticmethod
    def generate(kty="RSA", size=NODE_KEY_SIZE, **kwargs):
        return NodeKey(generate=kty, size=size, **kwargs)

    def __eq__(self, other):
        if isinstance(other, jwk.JWK):
            other_exported = other.export()
        else:
            other_exported = other
        return self.export() == other_exported

    def __repr__(self):
        return "<NodeKey {}>".format(self.thumbprint())


class DataKey(jwk.JWK):
    @staticmethod
    def generate(kty="oct", size=DATA_KEY_SIZE, **kwargs):
        return DataKey(generate=kty, size=size, **kwargs)

    def __repr__(self):
        return "<DataKey {}>".format(self.thumbprint())


class KeyChain:
    """
    Implementation of the keychain, which is distinct from the database model
    so that it can be kept up to date without relying on the sqlalchemy session
    mapper.
    """

    MAXIMUM_NODE_KEYS = 64

    def __init__(self, appsync_json=None):
        self.tag = ConfigurationObjectTag.KEY_CHAIN
        self.id = ConfigurationObjectTag.KEY_CHAIN.name
        self.serial = 0
        self.deleted = False
        self.update(appsync_json)

    def update(self, appsync_json):
        self._data_keys = {}
        self._node_keys = {}
        self._default_key_id = None
        self.serial = 0

        if appsync_json:
            configuration = self.decrypt(appsync_json["configuration"])

            for dkj in configuration["data_keys"]:
                dk = DataKey(**json.loads(dkj))
                self._data_keys[dk.thumbprint()] = dk

            for nkj in configuration["node_keys"]:
                nk = NodeKey(**json.loads(nkj))
                self._node_keys[nk.thumbprint()] = nk
            self._default_key_id = configuration["default_key_id"]
            self._serial = appsync_json["serial"]
            self.created_at = appsync_json.get("created_at")
            self.updated_at_at = appsync_json.get("updated_at_at")
            self.deleted = appsync_json.get("deleted", False)

    @property
    def serial(self):
        return self._serial

    @serial.setter
    def serial(self, value):
        self._serial = int(value)

    @property
    def data_keys(self):
        return [dk for dk in self._data_keys.values()]

    @property
    def node_keys(self):
        return [nk for nk in self._node_keys.values()]

    @property
    def default_key_id(self):
        return self._default_key_id

    @default_key_id.setter
    def default_key_id(self, value):
        if value not in self._data_keys:
            raise ValueError("Unknown data key")
        self._default_key_id = value

    def add_data_key(self, key=None, make_default=True):
        if key is None:
            key = DataKey.generate()
        elif isinstance(key, dict):
            new_key = DataKey()
            new_key.import_key(**key)
            key = new_key
        elif not isinstance(key, DataKey):
            raise ValueError("Invalid data key")

        tp = key.thumbprint()
        if tp in self._data_keys:
            raise ValueError("Key already exists")

        self._data_keys[tp] = key

        if make_default:
            self._default_key_id = tp
        return key

    def remove_data_key(self, key_id):
        if key_id in self._data_keys and len(self._data_keys.keys()) > 0:
            self._data_keys.pop(key_id, None)
            if self._default_key_id == key_id:
                self._default_key_id = self._data_keys.keys()[0]

    def add_node_key(self, key=None):
        if key is None:
            key = NodeKey.generate()
        elif isinstance(key, dict):
            new_key = NodeKey()
            new_key.import_key(**key)
            key = new_key
        elif not isinstance(key, NodeKey):
            raise ValueError("Invalid data key")

        tp = key.thumbprint()
        if tp in self._node_keys:
            raise ValueError("Key already exists", key)

        if (len(self._node_keys)) >= KeyChain.MAXIMUM_NODE_KEYS:
            raise LimitExceededError("Too many keys associated with endpoint")

        self._node_keys[tp] = key
        return key

    def remove_node_key(self, key_id):
        self._node_keys.pop(key_id, None)

    def decrypt(self, token):
        if token is None or token == "{}":
            return {}
        jwetoken = jwe.JWE()
        jwetoken.allowed_algs = [DATA_ALG, DATA_ENC, NODE_ALG, NODE_ENC]
        jwetoken.deserialize(token)
        header = jwetoken.jose_header

        key = None
        if header["alg"] == DATA_ALG and header["enc"] == DATA_ENC:
            kid = header["kid"]
            key = self._data_keys.get(kid)
            if not key:
                raise DataKeyError("Unknown key {}".format(kid))
            jwetoken.decrypt(key)
        elif header["alg"] == NODE_ALG and header["enc"] == NODE_ENC:
            decrypted = False
            for n in self.node_keys:
                try:
                    if not decrypted and n.has_private:
                        jwetoken.decrypt(n)
                        decrypted = True
                        break
                except Exception:
                    pass

            if not decrypted:
                if "DEPLOYMENT_KEY" in current_app.config:
                    jwetoken.decrypt(current_app.config["DEPLOYMENT_KEY"])
                elif "NODE_KEY" in current_app.config:
                    jwetoken.decrypt(current_app.config["NODE_KEY"])
                else:
                    raise DataKeyError("No deployment or node key available")
        else:
            raise ValueError("Invalid JWE header {}".format(header))

        return json.loads(jwetoken.payload.decode("utf-8"))

    def encrypt(self, token=None):
        if token and token.get("DATA_TYPE") != keychain_schema._DATA_TYPE:
            protected_header = {
                "alg": DATA_ALG,
                "enc": DATA_ENC,
                "typ": "JWE",
            }

            keys = [self._data_keys[self._default_key_id]]
            compact = True

            assert keys
        elif token:
            if self._node_keys:
                protected_header = {
                    "alg": NODE_ALG,
                    "enc": NODE_ENC,
                    "typ": "JWE",
                }
                keys = self.node_keys
                compact = False
            else:
                raise DataKeyError("No node keys available")
        else:
            return "{}"

        if token is None:
            token = {}

        jwetoken = jwe.JWE(
            json.dumps(token).encode("utf-8"), protected=protected_header
        )
        for k in keys:
            jwetoken.add_recipient(k, {"kid": k.thumbprint()})

        return jwetoken.serialize(compact=compact)

    @property
    def configuration(self):
        return self.encrypt(keychain_schema.dump(self))
