from __future__ import annotations

import ctypes
import ctypes.util
import os
from typing import Any, Dict, Optional

from distro import linux_distribution

from globus.manager.version import __version_fingerprint__ as version_fingerprint


class PlatformInfo:
    os_version = None
    openssl_version = None
    gcs_version = version_fingerprint

    def __init__(self) -> None:
        name = linux_distribution()[0].lower()
        self.platform = None
        platforms = {
            "redhat": [
                "almalinux",
                "centos",
                "fedora",
                "oracle",
                "red hat",
                "redhat",
                "rocky",
                "scientific",
                "springdale",
            ],
            "debian": ["debian", "ubuntu"],
            "suse": ["suse"],
        }
        forced = os.environ.get("GCS_FORCE_PLATFORM", "").lower()
        if forced in platforms.keys():
            self.platform = forced
        else:
            for p in platforms:
                if [x for x in platforms[p] if name.startswith(x)]:
                    self.platform = p
                    break

    def fingerprint(self) -> Dict[str, Any]:
        fingerprint = {
            "OS": PlatformInfo.__os_version(),
            "GCS": PlatformInfo.gcs_version,
        }
        openssl_version = PlatformInfo.__openssl_version()
        if openssl_version:
            fingerprint["OPENSSL"] = openssl_version
        return fingerprint

    @staticmethod
    def __os_version() -> Dict[str, Any]:
        if PlatformInfo.os_version is None:
            uname = os.uname()
            dist = linux_distribution(full_distribution_name=False)
            PlatformInfo.os_version = {
                "name": uname[0],
                "kernel": uname[2],
                "distribution": dist[0].lower(),
                "version": dist[1],
                "architecture": uname[4],
            }

        return PlatformInfo.os_version

    @staticmethod
    def __openssl_version() -> Optional[Dict[str, Any]]:
        if PlatformInfo.openssl_version is None:
            try:
                openssl_version_num = None
                libcrypto = ctypes.CDLL(ctypes.util.find_library("crypto"))
                if getattr(libcrypto, "SSLeay", None):
                    libcrypto.SSLeay.restype = ctypes.c_long
                    openssl_version_num = libcrypto.SSLeay()
                elif getattr(libcrypto, "OpenSSL_version_num", None):
                    libcrypto.OpenSSL_version_num.restype = ctypes.c_long
                    openssl_version_num = libcrypto.OpenSSL_version_num()
                if openssl_version_num is not None:
                    openssl_patch = (openssl_version_num >> 4) & 0xFF
                    openssl_status = openssl_version_num & 0xF
                    if openssl_status == 0:
                        openssl_status = "dev"
                    elif openssl_status < 0xF:
                        openssl_status = "beta-{}".format(openssl_status)
                    else:
                        openssl_status = "release"
                    PlatformInfo.openssl_version = {
                        "major": (openssl_version_num >> 28) & 0xF,
                        "minor": (openssl_version_num >> 20) & 0xFF,
                        "fix": (openssl_version_num >> 12) & 0xFF,
                        "patch": chr(ord("a") - 1 + openssl_patch),
                        "status": openssl_status,
                    }
            except Exception:
                pass
        return PlatformInfo.openssl_version


class UnsupportedPlatformException(Exception):
    def __init__(self, name: str | None = None):
        if name is None:
            name = PlatformInfo().platform
        super(UnsupportedPlatformException, self).__init__(
            "Unsupported platform {}".format(name)
        )
