import os
import copy
import yaml
import logging
from pathlib import Path
from shutil import copyfile
from yaml import Loader, SafeLoader
from collections.abc import Mapping, MutableMapping


def construct_yaml_str(self, node):
    return self.construct_scalar(node)


Loader.add_constructor('tag:yaml.org,2002:str', construct_yaml_str)
SafeLoader.add_constructor('tag:yaml.org,2002:str', construct_yaml_str)

log = logging.getLogger(__name__)


# Based on https://stackoverflow.com/a/9997519/845210
# Makes sure nobody tinkers with settings at runtime.
class ImmutableDict(Mapping):
    def __init__(self, somedict):
        selfdict = {}
        for k, v in somedict.items():
            if isinstance(v, MutableMapping):
                v = ImmutableDict(v)
            selfdict[k] = v
        self._dict = selfdict
        self._hash = None

    def __repr__(self):
        return 'ImmutableDict({})'.format(repr(self._dict))

    def __getitem__(self, key):
        return self._dict[key]

    def __len__(self):
        return len(self._dict)

    def __iter__(self):
        return iter(self._dict)

    def __hash__(self):
        if self._hash is None:
            self._hash = hash(frozenset(list(self._dict.items())))
        return self._hash

    def __eq__(self, other):
        return self._dict == other._dict

    def as_dict(self):
        d = copy.deepcopy(self._dict)
        for k, v in list(d.items()):
            if isinstance(v, ImmutableDict):
                d[k] = v.as_dict()
        return d


# Load static config from local, app-sync data from ~/globus-oidc
# Can be overwritten with the GLOBUS_OIDC_CONFIG_DIR environment variable
CONFIG_DIR = os.path.dirname(os.path.abspath(__file__))
INSTANCE_CONFIG_DIR = os.path.join(str(Path.home()), "globus-oidc")
if 'GLOBUS_OIDC_CONFIG_DIR' in os.environ:
    CONFIG_DIR = os.environ['GLOBUS_OIDC_CONFIG_DIR']
    INSTANCE_CONFIG_DIR = os.environ['GLOBUS_OIDC_CONFIG_DIR']


def load_config(config_dir: str, filename: str):
    conf_path = os.path.join(config_dir, filename)
    try:
        # resolve to path, open and read as a dict
        with open(conf_path) as f:
            return yaml.safe_load(f)
    except FileNotFoundError as ex:
        log.error("Unable to load configuration file '{}'".format(conf_path))
        raise ex


def initialize_page_templates(instance_config_dir: str):
    site_dir = os.path.join(instance_config_dir, "site")
    site_asset_dir = os.path.join(site_dir, "assets")
    source_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                  "..", "templates")
    asset_dir = os.path.join(source_dir, "assets")
    if not os.path.exists(site_dir):
        os.makedirs(site_dir)
        os.makedirs(site_asset_dir)
        copyfile(os.path.join(source_dir, 'login.mako'),
                 os.path.join(site_dir, 'login.mako'))
        for a in ['logo_here.png', 'styles.css']:
            copyfile(os.path.join(asset_dir, a),
                     os.path.join(site_asset_dir, a))


def validate_config(config: dict):
    """
    Fail early if configuration parameters will definitely cause this application
    to error or be unsafe.
    :param config:
    :return:
    """
    if config['oidc_server']['shared_key'] is None \
       or not config['oidc_server']['shared_key'].strip():
        raise RuntimeError("Config error: Shared key cannot be empty or whitespace")
    if config['oidc_server']['shared_session_key'] is None \
       or not config['oidc_server']['shared_session_key'].strip():
        raise RuntimeError("Config error: Shared session key cannot be empty or whitespace")
    if config['auth_client']['domain'] is None \
       or not config['auth_client']['domain'].strip():
        raise RuntimeError("Config error: Domain cannot be empty or whitespace")
    if config['oidc_server']['pam_service'] is None \
       or not config['oidc_server']['pam_service'].strip():
        raise RuntimeError("Config error: PAM service cannot be empty or whitespace")
    if len(config['clients']) == 0:
        raise RuntimeError("Config error: One client must be configured for the OIDC server")
    if config['FLASK_APP_DEBUG']:
        log.warning("Flask application is set to debug, not suitable for production!")


def settings():
    if settings._settings is None:
        # Combine both the configurations into the config key.
        config = load_config(CONFIG_DIR, "config.yaml")
        config.update(load_config(INSTANCE_CONFIG_DIR, "config_appsync.yaml"))
        # Update the key location to be the instance folder location.
        config["KEY_CONFIG"][0]["key"] = os.path.join(INSTANCE_CONFIG_DIR, "keys", "key.pem")
        # Validate the configuration
        validate_config(config)

        initialize_page_templates(INSTANCE_CONFIG_DIR)

        settings_dict = {
            "config": config,
            "logging": load_config(CONFIG_DIR, "logging.yaml"),
            "site": {
                "dir": os.path.join(INSTANCE_CONFIG_DIR, "site")
            }
        }
        settings._settings = ImmutableDict(settings_dict)
    return settings._settings


settings._settings = None
